/* $Id: main.h 643 2016-08-03 22:40:45Z bhockney $ */

/* (C) 2004-2016 by Bob Hockney <zeus@ix.netcom.com>                         *
 * Based on fwlogwatch written by                                            *
 * Boris Wesslowski <Wesslowski@CERT.Uni-Stuttgart.DE>                       *
 *                                                                           *
 * wfwl_syslog is the backend syslog parser for webfwlog.                    *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify      *
 * it under the terms of the GNU General Public License as published by      *
 * the Free Software Foundation; either version 2 of the License, or         *
 * (at your option) any later version.                                       *
 *                                                                           *
 * This program is distributed in the hope that it will be useful,           *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 * GNU General Public License for more details.                              *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to the Free Software               *
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#ifndef _MAIN_H
#define _MAIN_H

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <regex.h>

/* config.h */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STDBOOL_H
#include <stdbool.h>
#else
typedef char bool;
#define false 0
#define true  1
#endif

#ifdef HAVE_LIBZ
#include <zlib.h>
#endif

#ifdef HAVE_MAXMINDDB_H
#include <maxminddb.h>
#endif

#ifdef HAVE_DB_H
#include <db.h>
#endif

/* Data sizes */

#define BUFSIZE 1024
#define BUFSIZE_S "1023"
#define QUERYSIZE 4096
#define FILESIZE 256
#define TIMESIZE 40
#define SHORTLEN 128
#define SHORTLEN_S "127"
#define SHOSTLEN 33
#define SHOSTLEN_S "32"
#define IPLEN 16
#define IP6LEN 40
#define IP6LEN_S "39"
#define MAXSORTSIZE 24

#ifndef INET6_ADDRSTRLEN
#define INET6_ADDRSTRLEN 46
#endif

/* Files */

#define INFILE "/var/log/messages" /* Default input file */
#define REPORTDEF "-"

/* Database */

enum {
  DB_FAILURE,
  DB_SUCCESS
};

enum {
  QUERY_OK,
  QUERY_ERROR
};

/* Modes */

enum {
  LOG_SUMMARY,
  LOG_DETAIL
};

/* Parser */

#define PARSER_IPCHAINS 1
#define PARSER_NETFILTER 2
#define PARSER_IPFILTER 4
#define PARSER_WIN_XP 8
#define PARSER_IPFW 16
#define PARSER_CISCO_IOS 32
#define PARSER_CISCO_PIX 64
#define PARSER_SNORT 128
#define PARSER_NETSCREEN 256

enum {
  PARSE_OK,
  PARSE_ERROR,
  PARSE_NO_HIT,
  PARSE_WRONG_FORMAT,
  PARSE_EXCLUDED
};

enum {
  IN_ADDR_OK,
  IN_ADDR_ERROR
};

/* TCP flags */

#define TCP_SYN 1
#define TCP_ACK 2
#define TCP_FIN 4
#define TCP_RST 8
#define TCP_PSH 16
#define TCP_URG 32
#define TCP_OPTS_EXACT 64
#define TCP_FLAGS_MATCH 128

/* ipchains support */

#define IPCHAINS_DATE 1
#define IPCHAINS_DATA 2
#define IPCHAINS_IPS 4

/* netfilter support */

#define NF_DATE 1
#define NF_IN 2
#define NF_SRC 4
#define NF_DST 8
#define NF_PROTO 16
#define NF_SPT 32
#define NF_DPT 64
#define NF_TYPE 128

enum {
  NF_OPT_NOPREFIX,
  NF_OPT_PREFIX,
  NF_OPT_SRC,
  NF_OPT_DST
};

/* ipfilter support */

#define IPF_DATE 1
#define IPF_DATA 2
#define IPF_PROTO 4
#define IPF_SRC_IP 8
#define IPF_DST_IP 16
#define IPF_SRC_PORT 32
#define IPF_DST_PORT 64
#define IPF_TYPE 128
#define IPF_NO_HIT 256

#define IPF_OPT_NONE 1
#define IPF_OPT_COUNT 2
#define IPF_OPT_SRC 4
#define IPF_OPT_DST 8
#define IPF_OPT_RES 16
#define IPF_OPT_PORT 32
#define IPF_OPT_RPORT 64

/* ipfw support */

#define IPFW_DATE 1
#define IPFW_CHAIN 2
#define IPFW_PROTO 4
#define IPFW_IPS 8
#define IPFW_IF 16
#define IPFW_NO_HIT 32

enum {
  IPFW_OPT_NONE,
  IPFW_OPT_ICMP,
  IPFW_OPT_OTHER,
  IPFW_OPT_PORTS
};

/* Sorting */

#define SORT_INV_NO_SORT 35767

/* enumerated list of columns by which sorting may be done */
enum {
  SORT_ID = 1,
  SORT_SNAME,
  SORT_COUNT,
  SORT_START_TIME,
  SORT_END_TIME,
  SORT_PROTOCOL,
  SORT_SOURCEHOST,
  SORT_SOURCEPORT,
  SORT_DESTHOST,
  SORT_DESTPORT,
  SORT_ICMP_TYPE,
  SORT_ICMP_CODE,
  SORT_ICMP_ECHOID,
  SORT_ICMP_ECHOSEQ,
  SORT_ICMP_MTU,
  SORT_ICMP_GW,
  SORT_IP_ID,
  SORT_IP_IHL,
  SORT_IP_TOS,
  SORT_IP_TOTLEN,
  SORT_IP_TTL,
  SORT_IP6_FLOWLABEL,
  SORT_IP6_PAY_LEN,
  SORT_OOB_FAMILY,
  SORT_LOCAL_HOST,
  SORT_IN_IF,
  SORT_OUT_IF,
  SORT_FWMARK,
  SORT_PREFIX, /* same as CHAINLABEL */
  SORT_LOCAL_TIME,
  SORT_OOB_TIME,
  SORT_TIME_USEC,
  SORT_OOB_PROTOCOL,
  SORT_MAC_SADDR,
  SORT_MAC_DADDR,
  SORT_MAC,
  SORT_FRAGOFF,
  SORT_CSUM,
  SORT_AHESP_SPI,
  SORT_TCP_SEQ,
  SORT_TCP_ACKSEQ,
  SORT_TCP_WINDOW,
  SORT_UDP_LEN,
  SORT_TCP_URGP,
  SORT_TCP_OPTS,
  SORT_CONTINENT_CODE_SRC,
  SORT_CONTINENT_NAME_SRC,
  SORT_COUNTRY_ISO_CODE_SRC,
  SORT_COUNTRY_NAME_SRC,
  SORT_SUBDIVISION_1_ISO_CODE_SRC,
  SORT_SUBDIVISION_1_NAME_SRC,
  SORT_SUBDIVISION_2_ISO_CODE_SRC,
  SORT_SUBDIVISION_2_NAME_SRC,
  SORT_CITY_NAME_SRC,
  SORT_METRO_CODE_SRC,
  SORT_TIME_ZONE_SRC,
  SORT_ACCURACY_RADIUS_SRC,
  SORT_REGISTERED_COUNTRY_GEONAME_ID_SRC,
  SORT_REPRESENTED_COUNTRY_GEONAME_ID_SRC,
  SORT_IS_ANONYMOUS_PROXY_SRC,
  SORT_IS_SATELLITE_PROVIDER_SRC,
  SORT_POSTAL_CODE_SRC,
  SORT_LATITUDE_SRC,
  SORT_LONGITUDE_SRC,
  SORT_CONTINENT_CODE_DST,
  SORT_CONTINENT_NAME_DST,
  SORT_COUNTRY_ISO_CODE_DST,
  SORT_COUNTRY_NAME_DST,
  SORT_SUBDIVISION_1_ISO_CODE_DST,
  SORT_SUBDIVISION_1_NAME_DST,
  SORT_SUBDIVISION_2_ISO_CODE_DST,
  SORT_SUBDIVISION_2_NAME_DST,
  SORT_CITY_NAME_DST,
  SORT_METRO_CODE_DST,
  SORT_TIME_ZONE_DST,
  SORT_ACCURACY_RADIUS_DST,
  SORT_REGISTERED_COUNTRY_GEONAME_ID_DST,
  SORT_REPRESENTED_COUNTRY_GEONAME_ID_DST,
  SORT_IS_ANONYMOUS_PROXY_DST,
  SORT_IS_SATELLITE_PROVIDER_DST,
  SORT_POSTAL_CODE_DST,
  SORT_LATITUDE_DST,
  SORT_LONGITUDE_DST
};

enum {
  ORDER_ASCENDING,
  ORDER_DESCENDING
};

struct sort_order {
  int field;
  char keyname[SHORTLEN];
  int order;
  int position;
  struct sort_order *next;
};

/* enumerated list of possible fields for report */
enum {
  COL_ID=1,
  COL_SNAME,
  COL_COUNT,
  COL_LOCAL_HOST,
  COL_LOCAL_TIME,
  COL_START_TIME,
  COL_END_TIME,
  COL_LOG_PREFIX,
  COL_FWMARK,
  COL_OOB_PROTOCOL,
  COL_MAC_SADDR,
  COL_MAC_DADDR,
  COL_MAC,
  COL_IN_IF,
  COL_OUT_IF,
  COL_IP_PROTO,
  COL_SOURCEHOST,
  COL_SOURCEHOST_NAME,
  COL_SOURCEPORT,
  COL_SOURCE_SERVICE,
  COL_DESTHOST,
  COL_DESTHOST_NAME,
  COL_DESTPORT,
  COL_DEST_SERVICE,
  COL_IP_ID,
  COL_IP_CSUM,
  COL_IP_FRAGOFF,
  COL_IP_IHL,
  COL_IP_TOS,
  COL_IP_TTL,
  COL_IP6_FLOWLABEL,
  COL_IP6_PAY_LEN,
  COL_IP_LEN,
  COL_OOB_FAMILY,
  COL_ICMP_TYPE,
  COL_ICMP_CODE,
  COL_TCP_SEQ,
  COL_TCP_ACKSEQ,
  COL_TCP_WINDOW,
  COL_TCP_OPTIONS,
  COL_TCP_URGP,
  COL_UDP_LEN,
  COL_ICMP_ECHOID,
  COL_ICMP_ECHOSEQ,
  COL_ICMP_MTU,
  COL_ICMP_GATEWAY,
  COL_AHESP_SPI,
  COL_CONTINENT_CODE_SRC,
  COL_CONTINENT_NAME_SRC,
  COL_COUNTRY_ISO_CODE_SRC,
  COL_COUNTRY_NAME_SRC,
  COL_SUBDIVISION_1_ISO_CODE_SRC,
  COL_SUBDIVISION_1_NAME_SRC,
  COL_SUBDIVISION_2_ISO_CODE_SRC,
  COL_SUBDIVISION_2_NAME_SRC,
  COL_CITY_NAME_SRC,
  COL_METRO_CODE_SRC,
  COL_TIME_ZONE_SRC,
  COL_ACCURACY_RADIUS_SRC,
  COL_REGISTERED_COUNTRY_GEONAME_ID_SRC,
  COL_REPRESENTED_COUNTRY_GEONAME_ID_SRC,
  COL_IS_ANONYMOUS_PROXY_SRC,
  COL_IS_SATELLITE_PROVIDER_SRC,
  COL_POSTAL_CODE_SRC,
  COL_LATITUDE_SRC,
  COL_LONGITUDE_SRC,
  COL_CONTINENT_CODE_DST,
  COL_CONTINENT_NAME_DST,
  COL_COUNTRY_ISO_CODE_DST,
  COL_COUNTRY_NAME_DST,
  COL_SUBDIVISION_1_ISO_CODE_DST,
  COL_SUBDIVISION_1_NAME_DST,
  COL_SUBDIVISION_2_ISO_CODE_DST,
  COL_SUBDIVISION_2_NAME_DST,
  COL_CITY_NAME_DST,
  COL_METRO_CODE_DST,
  COL_TIME_ZONE_DST,
  COL_ACCURACY_RADIUS_DST,
  COL_REGISTERED_COUNTRY_GEONAME_ID_DST,
  COL_REPRESENTED_COUNTRY_GEONAME_ID_DST,
  COL_IS_ANONYMOUS_PROXY_DST,
  COL_IS_SATELLITE_PROVIDER_DST,
  COL_POSTAL_CODE_DST,
  COL_LATITUDE_DST,
  COL_LONGITUDE_DST
};

#define CACHE_RESOLVE 1
#define CACHE_POPULATE 2
#define CACHE_UPDATE 4
#define CACHE_POPULATE_FULL 8

struct field_order {
  int field;
  char keyname[SHORTLEN]; /* human readable for messages */
  int position;
  struct field_order *next;
};

/* Fields in report */
/* this is used to indicate which fields are in report for more efficient summarization */
struct fields_used {
  unsigned char count;
  unsigned char sname;
  unsigned char local_time;
  unsigned char earliest;
  unsigned char latest;
  unsigned char hostname;
  unsigned char log_label;
  unsigned char protocol;
  unsigned char totlen;
  unsigned char shost;
  unsigned char shost_name;
  unsigned char sport;
  unsigned char src_service;
  unsigned char dhost;
  unsigned char dhost_name;
  unsigned char dport;
  unsigned char dst_service;
  unsigned char flags;
  unsigned char oob_protocol;
  unsigned char mac_saddr_str;
  unsigned char mac_daddr_str;
  unsigned char raw_mac;
  unsigned char fwmark;
  unsigned char inif;
  unsigned char outif;
  unsigned char tos;
  unsigned char ttl;
  unsigned char ihl;
  unsigned char csum;
  unsigned char ipid;
  unsigned char family;
  unsigned char fragoff;
  unsigned char pay_len;
  unsigned char flowlabel;
  unsigned char tcp_seq;
  unsigned char tcp_ack_seq;
  unsigned char tcp_window;
  unsigned char tcp_urgp;
  unsigned char udp_len;
  unsigned char icmp_type;
  unsigned char icmp_code;
  unsigned char icmp_echoid;
  unsigned char icmp_echoseq;
  unsigned char icmp_gw;
  unsigned char icmp_mtu;
  unsigned char ahesp_spi;
  unsigned char continent_code_s;
  unsigned char continent_name_s;
  unsigned char country_iso_code_s;
  unsigned char country_name_s;
  unsigned char subdivision_1_iso_code_s;
  unsigned char subdivision_1_name_s;
  unsigned char subdivision_2_iso_code_s;
  unsigned char subdivision_2_name_s;
  unsigned char city_name_s;
  unsigned char metro_code_s;
  unsigned char time_zone_s;
  unsigned char accuracy_radius_s;
  unsigned char registered_country_geoname_id_s;
  unsigned char represented_country_geoname_id_s;
  unsigned char is_anonymous_proxy_s;
  unsigned char is_satellite_provider_s;
  unsigned char postal_code_s;
  unsigned char latitude_s;
  unsigned char longitude_s;
  unsigned char continent_code_d;
  unsigned char continent_name_d;
  unsigned char country_iso_code_d;
  unsigned char country_name_d;
  unsigned char subdivision_1_iso_code_d;
  unsigned char subdivision_1_name_d;
  unsigned char subdivision_2_iso_code_d;
  unsigned char subdivision_2_name_d;
  unsigned char city_name_d;
  unsigned char metro_code_d;
  unsigned char time_zone_d;
  unsigned char accuracy_radius_d;
  unsigned char registered_country_geoname_id_d;
  unsigned char represented_country_geoname_id_d;
  unsigned char is_anonymous_proxy_d;
  unsigned char is_satellite_provider_d;
  unsigned char postal_code_d;
  unsigned char latitude_d;
  unsigned char longitude_d;
};

/* Data structures */

/* Used for ICMP type and code tables */
struct icmp_code {
  int code;
  char *name;
};

struct icmp_type {
  int type;
  char *name;
  struct icmp_code *codes;
  int numcodes;
};

struct reportdef_keys {
  char * key;
  unsigned char sort_key;
  uint32_t sel_key;
  unsigned char sel_type;
  time_t * aggregate_value;
  unsigned char * field_used;
  unsigned char col_key;
  unsigned char cache_resolve;
  unsigned char use_geoip;
};

/* Geoip2 support */

#define GEOIP_USE_SRC 1
#define GEOIP_USE_DST 2
#define GEOIP_OFF     4

/* Keep order of this enum list in sync with struct geoip_data, and also *
 * array at top of parsers.c                                             */
enum {
  GEOIP2_CONT_CODE,
  GEOIP2_CONT_NAME,
  GEOIP2_COUNTRY_ISO,
  GEOIP2_COUNTRY_NAME,
  GEOIP2_SUB_0_ISO,
  GEOIP2_SUB_0_NAME,
  GEOIP2_SUB_1_ISO,
  GEOIP2_SUB_1_NAME,
  GEOIP2_CITY,
  GEOIP2_METRO,
  GEOIP2_TZ,
  GEOIP2_REG_COUNTRY_NAME,
  GEOIP2_REP_COUNTRY_NAME,
  GEOIP2_ANON_PROXY,
  GEOIP2_SAT_PROV,
  GEOIP2_POSTAL,
  GEOIP2_LATITUDE,
  GEOIP2_LONGITUDE,
  GEOIP2_ACCURACY_RADIUS,
  GEOIP2_MAX
};

struct geoip_data {
  char     continent_code[SHOSTLEN];
  char     continent_name[SHOSTLEN];
  char     country_iso_code[SHOSTLEN];
  char     country_name[SHOSTLEN];
  char     subdivision_1_iso_code[SHOSTLEN];
  char     subdivision_1_name[SHOSTLEN];
  char     subdivision_2_iso_code[SHOSTLEN];
  char     subdivision_2_name[SHOSTLEN];
  char     city_name[SHOSTLEN];
  uint16_t metro_code;
  char     time_zone[SHOSTLEN];
  char     registered_country_geoname_id[SHOSTLEN];
  char     represented_country_geoname_id[SHOSTLEN];
  bool     is_anonymous_proxy;
  bool     is_satellite_provider;
  char     postal_code[SHOSTLEN];
  double   latitude;
  double   longitude;
  uint16_t accuracy_radius;
  uint32_t refcount;
};

/* Used for receiving data from a line in log file */
struct log_line {
  char filename[FILESIZE];
  unsigned long int linenum;
  time_t time;
  char hostname[SHOSTLEN];
  char log_label[SHORTLEN];
  char chainlabel[SHORTLEN];
  char branchname[SHORTLEN];
  char interface[SHORTLEN];
  unsigned char protocol;
  uint16_t totlen;
  unsigned short int family;
  struct in_addr shost;
#ifdef HAVE_INET_NTOP
  struct in6_addr shost6;
#endif
  char shost_s[INET6_ADDRSTRLEN];
  char shostname[SHORTLEN];
  uint32_t sport;
  char sservice[SHORTLEN];
  struct in_addr dhost;
#ifdef HAVE_INET_NTOP
  struct in6_addr dhost6;
#endif
  char dhost_s[INET6_ADDRSTRLEN];
  char dhostname[SHORTLEN];
  uint32_t dport;
  char dservice[SHORTLEN];
  unsigned char flags;
  unsigned long int count;
  uint16_t oob_protocol; 
  char mac_saddr_str[SHORTLEN];
  char mac_daddr_str[SHORTLEN];
  char raw_mac[SHORTLEN];
  uint32_t fwmark;
  char inif[SHORTLEN];
  char outif[SHORTLEN];
  unsigned char tos;
  unsigned char ttl;
  unsigned char ihl;
  uint16_t csum;
  uint16_t ipid;
  uint16_t pay_len;
  uint32_t flowlabel;
  uint16_t fragoff;
  uint32_t tcp_seq;
  uint32_t tcp_ack_seq;
  uint16_t tcp_window;
  uint16_t tcp_urgp;
  uint16_t udp_len;
  unsigned char icmp_type;
  unsigned char icmp_code;
  uint16_t icmp_echoid;
  uint16_t icmp_echoseq;
  struct in_addr icmp_gw;
  uint16_t icmp_mtu;
  uint32_t ahesp_spi;
  struct geoip_data * geoip_data_src;
  struct geoip_data * geoip_data_dst;
};

/* Linked list of report data */
struct conn_data {
  char filename[FILESIZE];
  unsigned long int rowid;
  unsigned long int linenum;
  unsigned long int count;
  time_t local_time;
  time_t start_time;
  time_t end_time;
  char hostname[SHOSTLEN];
  char log_label[SHORTLEN];
  char chainlabel[SHORTLEN];
  char branchname[SHORTLEN];
  char interface[SHORTLEN];
  unsigned char protocol;
  uint16_t totlen;
  unsigned short int family;
  struct in_addr shost;
#ifdef HAVE_INET_NTOP
  struct in6_addr shost6;
#endif
  char shost_s[INET6_ADDRSTRLEN];
  char shostname[SHORTLEN];
  uint32_t sport;
  char sservice[SHORTLEN];
  struct in_addr dhost;
#ifdef HAVE_INET_NTOP
  struct in6_addr dhost6;
#endif
  char dhost_s[INET6_ADDRSTRLEN];
  char dhostname[SHORTLEN];
  uint32_t dport;
  char dservice[SHORTLEN];
  unsigned char flags;
  uint16_t oob_protocol;
  char mac_saddr_str[SHORTLEN];
  char mac_daddr_str[SHORTLEN];
  char raw_mac[SHORTLEN];
  uint32_t fwmark;
  char inif[SHORTLEN];
  char outif[SHORTLEN];
  unsigned char tos;
  unsigned char ttl;
  unsigned char ihl;
  uint16_t csum;
  uint16_t ipid;
  uint16_t pay_len;
  uint32_t flowlabel;
  uint16_t fragoff;
  uint32_t tcp_seq;
  uint32_t tcp_ack_seq;
  uint16_t tcp_window;
  uint16_t tcp_urgp;
  uint16_t udp_len;
  unsigned char icmp_type;
  unsigned char icmp_code;
  uint16_t icmp_echoid;
  uint16_t icmp_echoseq;
  struct in_addr icmp_gw;
  uint16_t icmp_mtu;
  uint32_t ahesp_spi;
  struct geoip_data * geoip_data_src;
  struct geoip_data * geoip_data_dst;
  struct conn_data *next;
};

/* when populating cache for summarized report at runtime, this structure */
/* contains detail for summarized lines for fields that may not appear in */
/* report */
struct conn_data_sum {
  unsigned long int rowid;
  unsigned short int family;
  struct in_addr shost;
#ifdef HAVE_INET_NTOP
  struct in6_addr shost6;
#endif
  char shost_s[INET6_ADDRSTRLEN];
  struct in_addr dhost;
#ifdef HAVE_INET_NTOP
  struct in6_addr dhost6;
#endif
  char dhost_s[INET6_ADDRSTRLEN];
  unsigned char protocol;
  uint32_t sport;
  uint32_t dport;
  struct conn_data_sum *next;
};

/* Linked list of files to be parsed */
struct input_file {
  char *name;
  struct input_file *next;
};

/* enumerated list of matches */
enum {
  P_MATCH_NONE,
  P_MATCH_EXC
};

/* include/exclude bitmap for numeric criteria, *
 * to support use of multiple values.           */
#define MATCH_MUL_NONE             0x000000000000
#define MATCH_TCP_DPORT            0x000000000001
#define MATCH_TCP_SPORT            0x000000000002
#define MATCH_UDP_DPORT            0x000000000004
#define MATCH_UDP_SPORT            0x000000000008
#define MATCH_ICMP_TYPE            0x000000000010
#define MATCH_ICMP_CODE            0x000000000020
#define MATCH_SRC_HOST             0x000000000040
#define MATCH_DST_HOST             0x000000000080
#define MATCH_PROTOCOL             0x000000000100
#define MATCH_AHESP_SPI            0x000000000200
#define MATCH_IP_TOS               0x000000000400
#define MATCH_IP_TTL               0x000000000800
#define MATCH_IP_IHL               0x000000001000
#define MATCH_IP_TOTLEN            0x000000002000
#define MATCH_IP_ID                0x000000004000
#define MATCH_IP_CSUM              0x000000008000
#define MATCH_IP_FRAGOFF           0x000000010000
#define MATCH_OOB_MARK             0x000000020000
#define MATCH_TCP_SEQ              0x000000040000
#define MATCH_TCP_ACKSEQ           0x000000080000
#define MATCH_TCP_WINDOW           0x000000100000
#define MATCH_TCP_URGP             0x000000200000
#define MATCH_ICMP_ECHOID          0x000000400000
#define MATCH_ICMP_ECHOSEQ         0x000000800000
#define MATCH_ICMP_FRAGMTU         0x000001000000
#define MATCH_UDP_LEN              0x000002000000
#define MATCH_ICMP_GATEWAY         0x000004000000
#define MATCH_OOB_FAMILY           0x000008000000
#define MATCH_ICMPv6_TYPE          0x000010000000
#define MATCH_ICMPv6_CODE          0x000020000000
#define MATCH_ICMPv6_ECHOID        0x000040000000
#define MATCH_ICMPv6_ECHOSEQ       0x000080000000
#define MATCH_IP6_FLOWLABEL        0x000100000000ULL
#define MATCH_IP6_PAY_LEN          0x000200000000ULL
#define MATCH_OOB_PROTOCOL         0x000400000000ULL
#define MATCH_METRO_CODE_SRC       0x000800000000ULL
#define MATCH_ACCURACY_RADIUS_SRC  0x001000000000ULL 
#define MATCH_LATITUDE_SRC         0x002000000000ULL
#define MATCH_LONGITUDE_SRC        0x004000000000ULL
#define MATCH_METRO_CODE_DST       0x008000000000ULL
#define MATCH_ACCURACY_RADIUS_DST  0x010000000000ULL 
#define MATCH_LATITUDE_DST         0x020000000000ULL
#define MATCH_LONGITUDE_DST        0x040000000000ULL

/* enumerated list of item types in report definitions */
enum {
  DEF_NONE,
  DEF_WHERE,
  DEF_INVERT,
  DEF_SORT,
  DEF_ORDER,
  DEF_COLUMN,
  DEF_SUM,
  DEF_OPTION
};

/* enumerated list of items by which selection may be done */
enum {
  SEL_MIN_DATE = 1,
  SEL_MAX_DATE,
  SEL_PROTOCOL,
  SEL_SOURCEHOST,
  SEL_DESTHOST,
  SEL_UDP_SPORT,
  SEL_UDP_DPORT,
  SEL_ICMP_TYPE,
  SEL_ICMP_CODE,
  SEL_ICMPv6_TYPE,
  SEL_ICMPv6_CODE,
  SEL_LOCAL_HOST,
  SEL_IN_IF,
  SEL_OUT_IF,
  SEL_PREFIX,
  SEL_TCP_SPORT,
  SEL_TCP_DPORT,
  SEL_TCP_FLAGS,
  SEL_TCP_SYN,
  SEL_TCP_ACK,
  SEL_TCP_FIN,
  SEL_TCP_RST,
  SEL_TCP_PSH,
  SEL_TCP_URG,
  SEL_TCP_OPTS_EXACT,
  SEL_AHESP_SPI,
  SEL_IP_TOS,
  SEL_IP_TTL,
  SEL_IP_IHL,
  SEL_IP_TOTLEN,
  SEL_IP_ID,
  SEL_IP_CSUM,
  SEL_IP_DF,
  SEL_IP_MF,
  SEL_IP_FRAGOFF,
  SEL_IP6_FLOWLABEL,
  SEL_IP6_PAY_LEN,
  SEL_OOB_FAMILY,
  SEL_OOB_MARK,
  SEL_TCP_SEQ,
  SEL_TCP_ACKSEQ,
  SEL_TCP_WINDOW,
  SEL_TCP_URGP,
  SEL_ICMP_ECHOID,
  SEL_ICMP_ECHOSEQ,
  SEL_ICMP_GATEWAY,
  SEL_ICMP_FRAGMTU,
  SEL_ICMPv6_ECHOID,
  SEL_ICMPv6_ECHOSEQ,
  SEL_UDP_LEN,
  SEL_OOB_PROTOCOL,
  SEL_MAC_SADDR,
  SEL_MAC_DADDR,
  SEL_MAC,
  SEL_CONTINENT_CODE_SRC,
  SEL_CONTINENT_NAME_SRC,
  SEL_COUNTRY_ISO_CODE_SRC,
  SEL_COUNTRY_NAME_SRC,
  SEL_SUBDIVISION_1_ISO_CODE_SRC,
  SEL_SUBDIVISION_1_NAME_SRC,
  SEL_SUBDIVISION_2_ISO_CODE_SRC,
  SEL_SUBDIVISION_2_NAME_SRC,
  SEL_CITY_NAME_SRC,
  SEL_METRO_CODE_SRC,
  SEL_TIME_ZONE_SRC,
  SEL_ACCURACY_RADIUS_SRC,
  SEL_REGISTERED_COUNTRY_GEONAME_ID_SRC,
  SEL_REPRESENTED_COUNTRY_GEONAME_ID_SRC,
  SEL_IS_ANONYMOUS_PROXY_SRC,
  SEL_IS_SATELLITE_PROVIDER_SRC,
  SEL_POSTAL_CODE_SRC,
  SEL_LATITUDE_SRC,
  SEL_LONGITUDE_SRC,
  SEL_CONTINENT_CODE_DST,
  SEL_CONTINENT_NAME_DST,
  SEL_COUNTRY_ISO_CODE_DST,
  SEL_COUNTRY_NAME_DST,
  SEL_SUBDIVISION_1_ISO_CODE_DST,
  SEL_SUBDIVISION_1_NAME_DST,
  SEL_SUBDIVISION_2_ISO_CODE_DST,
  SEL_SUBDIVISION_2_NAME_DST,
  SEL_CITY_NAME_DST,
  SEL_METRO_CODE_DST,
  SEL_TIME_ZONE_DST,
  SEL_ACCURACY_RADIUS_DST,
  SEL_REGISTERED_COUNTRY_GEONAME_ID_DST,
  SEL_REPRESENTED_COUNTRY_GEONAME_ID_DST,
  SEL_IS_ANONYMOUS_PROXY_DST,
  SEL_IS_SATELLITE_PROVIDER_DST,
  SEL_POSTAL_CODE_DST,
  SEL_LATITUDE_DST,
  SEL_LONGITUDE_DST
};

/* enumerated list of selection criteria types */
enum {
  SEL_NUM = 1,
  SEL_DOUBLE,
  SEL_REGEX,
  SEL_BOOLEAN,
  SEL_IPADDR,
  SEL_PROTO,
  SEL_DATE,
  SEL_DFMF,
  SEL_AGG_COUNT,
  SEL_AGG_DATE,
  SEL_ID
};

struct selection {
  int field;
  char keyname[SHORTLEN];
  int type;
  unsigned long int value;
  unsigned long int max_value;
  double double_value;
  double max_double_value;
  char svalue[SHORTLEN];
  regex_t buf;
  unsigned short int family;
  uint32_t in_addr;
  uint32_t netmask;
#ifdef HAVE_INET_NTOP
  struct in6_addr in6_addr;
#endif
  unsigned char in6_bits;
  unsigned char invert;
  int have_value;
  struct selection *next;
};

/* Holds values for aggregate selection criteria */
struct select_sum {
  unsigned long int min_count;
  unsigned long int max_count;
  time_t max_earliest;
  time_t min_latest;
};

/* Verbosity levels.                     *
 * The labels are for convenience only.  */
enum {
  VERBOSE_OFF,     /* No status messages */
  VERBOSE_INFO,    /* Basic status messages of actions being taken */
  VERBOSE_NOTICE,  /* Detail status messages */
  VERBOSE_ALERT,   /* Still more status messages */
  VERBOSE_WARNING, /* Some detail of progress */
  VERBOSE_ERROR,   /* More detail of progress */
  VERBOSE_DEBUG    /* Great detail of progress */
};

/* Contains parsed global options. */
/* Also contains the current log line being parsed and other reused data */
/* There is one of these with global scope */
struct options {
  unsigned char mode;           /* reporting mode */
#ifdef HAVE_LIBZ
  gzFile inputfd;               /* file handle of current input file */
#else
  FILE * inputfd;               /* file handle of current input file */
#endif
  char pathname[FILESIZE];      /* path to syslog files */
  char filename[FILESIZE];      /* name of file currently being parsed */
  unsigned long int packet;     /* set to line number for packet detail */
  unsigned long int linenum;    /* line number currently being parsed */
  unsigned char std_in;         /* set if no input files; reads stdin */
  char geoip2_locale[SHORTLEN]; /* currently used only for geoip2 data */

  unsigned char verbose;        /* verbosity level: 0 to 5 */
  int filecount;                /* number of input files, or 0 if stdin used */
  char reportdef[FILESIZE];     /* name of report definition file */
  char db[SHORTLEN];            /* database to use - mysql or pgsql */
  char mysql_server[SHORTLEN];  /* location of mysql server */
  char mysql_user[SHORTLEN];    /* mysql user name */
  char mysql_pass[SHORTLEN];    /* mysql password */
  char mysql_wfwl_db[SHORTLEN]; /* mysql database */
  unsigned char have_db;        /* set if database support is compiled in */

  char pgsql_server[SHORTLEN];  /* location of postgresql server */
  char pgsql_user[SHORTLEN];    /* postgresql user name */
  char pgsql_pass[SHORTLEN];    /* postgresql password */
  char pgsql_db[SHORTLEN];      /* postgresql database */
  char pgsql_wfwl_schema[SHORTLEN]; /* postgresql schema */
  unsigned char pgsql_have_namespace; /* set if server supports schemas */
 
  struct log_line *line;        /* data for current log line being parsed */
  char format_sel[SHORTLEN];    /* list of parsers from command line */
  unsigned int format;          /* list of parsers to use */
  unsigned int parser;          /* used internally by parser to determine what has been parsed */
  unsigned char repeated;       /* used for repeat expansion */
  int orig_count;               /* 0 or 1; used with repeat expansion */
  unsigned char geoip_version;  /* geoip version to use, 1 or 2 */
  char geoip2_file[BUFSIZE];    /* geoip filename */
#ifdef HAVE_MAXMINDDB_H
  MMDB_s *geoip2_handle;        /* geoip database handle */
#endif
  unsigned char use_geoip;      /* bitmap for whether source, destination, or neither fields are used */
  struct geoip_data *geoip_data_init; /* initialized blank geoip_data struct to initialize lines */
  bool use_db;                  /* whether or not to use berkeley db with geoip */
#ifdef HAVE_DB_H
  DB *dbp;                      /* pointer to berkeley db */
#endif

  unsigned char sortfield;      /* field to sort by */
  unsigned char sortmode;       /* ascending or descending */
  unsigned char resolve_hosts;  /* set if hostnames and service names are to be resolved */

  unsigned long int report_rows; /* number of total rows in report */
  unsigned long int matched_entries; /* number of log entries meeting selection criteria */
  unsigned char use_out;        /* set if outfile to be used */
  char outputfile[FILESIZE];    /* name of outfile */

  unsigned long int max;        /* maximum number of rows to output */
  unsigned long int begin;      /* starting row to output */

  unsigned char ipchains_check; /* set to check ipchains rules */
};

#endif
