/* $Id: report.c 643 2016-08-03 22:40:45Z bhockney $ */

/* (C) 2004-2016 by Bob Hockney <zeus@ix.netcom.com>    *
 * Based on fwlogwatch written by                       *
 * Boris Wesslowski <Wesslowski@CERT.Uni-Stuttgart.DE>  *
 *                                                      *
 * report generation for wfwl_syslog                    *
 *                                                      *
 * This code is distributed under the terms of GNU GPL  */

#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <syslog.h>
#include <signal.h>

#include <main.h>
#include <parser.h>
#include <compare.h>
#include <utils.h>
#include <database.h>
#include <output.h>
#include <getdate.h>

#ifdef HAVE_LIBZ
#include <zlib.h>
#endif

#ifdef HAVE_DB_H
extern DBT key, data;
#endif

/* These are structs have global scope */
extern struct options opt;
extern struct conn_data *first;
extern struct input_file *first_file;
extern struct field_order *fields;
extern struct sort_order *sort;
extern struct fields_used fields_used;
extern struct select_sum select_sum;

/* select based on aggregate criteria */
struct conn_data * select_aggregate(struct conn_data *list) {
  struct conn_data *this, *new=NULL, *add, *old;
  unsigned long int numrows = 0;

  this = list;

  while (this != NULL) {
    if (!((select_sum.min_count != 0 && this->count < select_sum.min_count) ||
          (select_sum.max_count != 0 && this->count > select_sum.max_count) ||
          (select_sum.max_earliest != 0 && this->start_time > select_sum.max_earliest) ||
          (select_sum.min_latest != 0 && this->end_time < select_sum.min_latest))) {
      numrows++;
      add = xmalloc(sizeof(struct conn_data));
      *add = *this;
      if (new == NULL) {
        new = add;
        new->next=NULL;
      } else {
        add->next = new;
        new = add;
      }
    }
    old = this;
    this = this->next;
    free(old);
  }

  opt.report_rows = numrows;

  return new;

}

/* PARSE INPUT FILES AND OUTPUT REPORT */
/* This function is the main routine for parsing and outputting reports */
/* It calls other functions as needed */
/* Called without parameters; uses global structs, above */
void generate_report()
{
  char buf[BUFSIZE], *input = NULL, last_file = 0;
  FILE *output = NULL;
  unsigned int j;
  int retval, linenum = 0, hit = 0;
  struct input_file *file;
  struct field_order *flds;
  struct sort_order *srt;
  struct timeval *segtime;
  int64_t hostsupdtime = 0, servicesupdtime = 0, populatecachetime = 0;
  struct geoip_data *geoip_data_init;
#ifdef HAVE_DB_H  
  char dbfile[L_tmpnam];
  int tmp, ret;
  int x = 0;
  int y = 0;
  DB *dbp;
  DBC *dbcp;
#endif

  opt.line = xmalloc(sizeof(struct log_line));

  if (opt.verbose >= VERBOSE_ERROR) {
    fprintf(stderr, ". = accepted\n");
    fprintf(stderr, "e = excluded\n");
    fprintf(stderr, "r = repeated\n");
    fprintf(stderr, "_ = not packet log entry\n");
  }

#ifdef HAVE_MAXMINDDB_H
  int status;
  MMDB_s mmdb;

  status = MMDB_open(opt.geoip2_file, MMDB_MODE_MMAP, &mmdb);

  if (MMDB_SUCCESS != status) {
    if(opt.verbose >= VERBOSE_NOTICE) {
      fprintf(stderr, "Can't open mmdb file %s, skipping geoip lookup\n  MMDB error is: %s\n", opt.geoip2_file, MMDB_strerror(status));

      if (MMDB_IO_ERROR == status) {
        fprintf(stderr, "  IO error: %s\n", strerror(errno));
      }
    }
  } else { /* success */
    opt.geoip2_handle = &mmdb;
    if (opt.verbose >= VERBOSE_ALERT)
      fprintf(stderr, "Opened mmdb file %s\n", opt.geoip2_file);
  }
#else
   if(opt.use_geoip < GEOIP_OFF && opt.use_geoip && opt.verbose >= VERBOSE_INFO)
    fprintf(stderr, "Geoip columns, sorts or criteria specified but no libmaxminddb library available. Is support compiled it?\n"); 
#endif

  /* Initialize single struct for geoip data to initialize *
   * geoip fields; only changed if data exists for ip      */
  geoip_data_init = xmalloc(sizeof(struct geoip_data));

  geoip_data_init->continent_code[0] = '\0';
  geoip_data_init->continent_name[0] = '\0';
  geoip_data_init->country_iso_code[0] = '\0';
  geoip_data_init->country_name[0] = '\0';
  geoip_data_init->subdivision_1_iso_code[0] = '\0';
  geoip_data_init->subdivision_1_name[0] = '\0';
  geoip_data_init->subdivision_2_iso_code[0] = '\0';
  geoip_data_init->subdivision_2_name[0] = '\0';
  geoip_data_init->city_name[0] = '\0';
  geoip_data_init->metro_code = 0;
  geoip_data_init->time_zone[0] = '\0';
  geoip_data_init->registered_country_geoname_id[0] = '\0';
  geoip_data_init->represented_country_geoname_id[0] = '\0';
  geoip_data_init->is_anonymous_proxy = false;
  geoip_data_init->is_satellite_provider = false;
  geoip_data_init->postal_code[0] = '\0';
  geoip_data_init->latitude = 0;
  geoip_data_init->longitude = 0;
  geoip_data_init->accuracy_radius = 0;
  geoip_data_init->refcount = 1;

  opt.geoip_data_init = geoip_data_init;

#ifdef HAVE_DB_H
  xstrncpy((char *)&dbfile,P_tmpdir,L_tmpnam);
  strncat((char *)&dbfile,"/wfwl-XXXXXX",L_tmpnam - strlen((char *)&dbfile) - 1);
  tmp = mkstemp((char *)&dbfile);

  if (tmp == -1)
    exit(0);
  else
    if (opt.verbose >= VERBOSE_DEBUG)
      fprintf(stderr, "tmp filename %s\n", (char *)&dbfile);

  close(tmp);
  unlink((char *)&dbfile);

  if ((ret = db_create(&dbp, NULL, 0)) != 0) {
    if (opt.verbose >= VERBOSE_INFO)
      fprintf(stderr, "db_create: %s\n", db_strerror(ret));
    goto err;
  }
  if ((ret = dbp->open(dbp, NULL, (char *)&dbfile, NULL, DB_BTREE, DB_CREATE, 0600)) != 0) {
    if (opt.verbose >= VERBOSE_INFO)
      fprintf(stderr, "db_open: %s\n", db_strerror(ret));
    goto err;
  }
  if (opt.verbose >= VERBOSE_ALERT)
    fprintf(stderr, "Berkeley db opened\n");

  opt.use_db = true;
  opt.dbp = dbp;
  memset(&key, 0, sizeof(key));
  memset(&data, 0, sizeof(data));

  err:

#endif
  /* get file handle of next file to process, stdin if '-' specified */
  file = first_file;
  while (last_file == 0) {
    if (opt.std_in) {
      if (opt.verbose >= VERBOSE_INFO)
        fprintf(stderr, "Processing input from stdin\n");

#ifdef HAVE_LIBZ
      opt.inputfd = (gzFile) stdin;
#else
      opt.inputfd = stdin;
#endif
    } else {
      input = file->name;
      if (opt.verbose >= VERBOSE_INFO)
        fprintf(stderr, "Processing input file '%s'\n", input);

#ifdef HAVE_LIBZ
      opt.inputfd = gzopen(input, "rb");
#else
      opt.inputfd = fopen(input, "r");
#endif
      if (opt.inputfd == NULL) {
#ifdef HAVE_LIBZ
        fprintf(stderr, "gzopen %s: %s\n", input, strerror(errno));
#else
        fprintf(stderr, "fopen %s: %s\n", input, strerror(errno));
#endif
      }
    }

    /* read first line from input file */
#ifdef HAVE_LIBZ
    if (opt.std_in) {
#endif
      retval = (fgets(buf, BUFSIZE, (FILE *)opt.inputfd) != NULL);
#ifdef HAVE_LIBZ
    } else {
      retval = (gzgets(opt.inputfd, buf, BUFSIZE) != Z_NULL);
    }
#endif

    xstrncpy(opt.filename, input + strlen(opt.pathname), FILESIZE);
    linenum = 0;

    if(opt.packet) {
      for (j = 1; j < opt.packet; j++) {
#ifdef HAVE_LIBZ
          if (opt.std_in) {
#endif
          retval = (fgets(buf, BUFSIZE, (FILE *)opt.inputfd) != NULL);
#ifdef HAVE_LIBZ
          } else {
            retval = (gzgets(opt.inputfd, buf, BUFSIZE) != Z_NULL);
          }
#endif
      }
    }

    /* process lines until EOF */
    while (retval) {
      ++linenum;
      hit = PARSE_NO_HIT;
      hit = parse_line(buf, linenum);
      opt.repeated = 0;
      switch (hit) {
        case PARSE_OK:
          opt.matched_entries++;
          opt.repeated = 1;
            break;
        case PARSE_WRONG_FORMAT:
        case PARSE_EXCLUDED:
            break;
        default:
            break;
      }

      /* only do one line if packet report */
      if (opt.packet) break;

      /* get next line from input file */
#ifdef HAVE_LIBZ
      if (opt.std_in) {
#endif
        retval = (fgets(buf, BUFSIZE, (FILE *)opt.inputfd) != NULL);
#ifdef HAVE_LIBZ
      } else {
        retval = (gzgets(opt.inputfd, buf, BUFSIZE) != Z_NULL);
      }
#endif
    }

    /* After each input file, print line break following progress indication */
    if (opt.verbose >= VERBOSE_ERROR)
      fprintf(stderr, ".\n");

    /* set next input file or indicate done with last_file=1 */
    if(opt.std_in) {
      last_file++;
    } else {
      if (opt.verbose >= VERBOSE_WARNING)
        fprintf(stderr, "Done processing input from '%s'\n", input);

#ifndef HAVE_LIBZ
      retval = fclose(opt.inputfd);
      if (retval == EOF) {
        perror("fclose");
#else
      retval = gzclose(opt.inputfd);
      if (retval != 0) {
        if (retval != Z_ERRNO) {
          fprintf(stderr, "gzclose %s: %s\n", input, gzerror(opt.inputfd, &retval));
        } else {
          perror("gzclose");
        }
#endif
      }

      if (file->next != NULL) {
        file = file->next;
      } else {
        last_file++;
      }
    }
  } /* end while (last_file == 0) */

#ifdef HAVE_MAXMINDDB_H
  if (opt.geoip2_handle !=NULL)
    MMDB_close(&mmdb);
  if(opt.verbose >= VERBOSE_ALERT)
    fprintf(stderr, "Closed mmdb file %s\n", opt.geoip2_file); 
#endif
#ifdef HAVE_DB_H
  /* Acquire a cursor for the database. */
  ret = dbp->cursor(dbp, NULL, &dbcp, 0);
  if (ret !=0 && opt.verbose >= VERBOSE_INFO)
    fprintf(stderr, db_strerror(ret));

  /* Re-initialize the key/data pair. */
  memset(&key, 0, sizeof(key));
  memset(&data, 0, sizeof(data));

  while ((ret = dbcp->c_get(dbcp, &key, &data, DB_NEXT)) == 0) {
      y++;
    if ((*(struct geoip_data **)data.data)->refcount == 0) {
      free(*(struct geoip_data **)data.data);
      x++;
    }
  }
  if(opt.verbose >= VERBOSE_ALERT)
    fprintf(stderr, "berkeley db record count: %u, unused %u\n", y, x);

  /* Close the cursor. */
  ret = dbcp->c_close(dbcp);
  if (ret !=0 && opt.verbose >= VERBOSE_INFO)
    fprintf(stderr, db_strerror(ret));

  ret = dbp->close(dbp, 0);
  if (ret !=0 && opt.verbose >= VERBOSE_INFO)
    fprintf(stderr, db_strerror(ret));

  if (opt.verbose >= VERBOSE_ALERT)
    fprintf(stderr, "db closed\n");

  unlink((char *)&dbfile);
#endif

  free(opt.line);

  if (opt.mode == LOG_SUMMARY)
    first = select_aggregate(first); 

  if (opt.verbose >= VERBOSE_INFO) {
    fprintf(stderr, "Sorting data\n");
  }

  srt = sort;
 
  while (srt != NULL && srt->position != SORT_INV_NO_SORT) {
    opt.sortfield = srt->field;
    opt.sortmode = srt->order;
    first = wfwl_sort(first);
    if(opt.verbose >= VERBOSE_ERROR)
      fprintf(stderr, ".");
    srt = srt->next;
  }

  if (opt.verbose >= VERBOSE_ERROR)
    fprintf(stderr, ".\n");

  segtime = xmalloc(sizeof(struct timeval));

  /* resolve host and service names if specified */
  if (opt.resolve_hosts & CACHE_POPULATE || opt.resolve_hosts & CACHE_UPDATE) {
    if (fields_used.dhost_name || fields_used.shost_name) {
      gettimeofday(segtime, NULL);
      if ((opt.resolve_hosts & CACHE_UPDATE))
        hostsupdtime -= segtime->tv_sec;
      else
        populatecachetime -= segtime->tv_sec;
      resolve_hosts();
      gettimeofday(segtime, NULL);
      if (opt.resolve_hosts & CACHE_UPDATE)
        hostsupdtime += segtime->tv_sec;
      else
        populatecachetime += segtime->tv_sec;
    }
    if (fields_used.dst_service || fields_used.src_service) {
      gettimeofday(segtime, NULL);
      if (opt.resolve_hosts & CACHE_UPDATE)
        servicesupdtime -= segtime->tv_sec;
      else
        populatecachetime -= segtime->tv_sec;
      resolve_services();
      gettimeofday(segtime, NULL);
      if (opt.resolve_hosts & CACHE_UPDATE)
        servicesupdtime += segtime->tv_sec;
      else
        populatecachetime += segtime->tv_sec;
    }
  }
  if (opt.resolve_hosts & CACHE_POPULATE_FULL) {
    gettimeofday(segtime, NULL);
    populatecachetime -= segtime->tv_sec;
    populate_cache();
    gettimeofday(segtime, NULL);
    populatecachetime += segtime->tv_sec;
  }

  /* use outfile if specified */
  if (opt.use_out) {
    if (opt.verbose >= VERBOSE_INFO)
      fprintf(stderr, "Writing to output file '%s'\n", opt.outputfile);

    /* redirect stdout to outfile */
    output = freopen(opt.outputfile, "w", stdout);
    if (output == NULL) {
      fprintf(stderr, "freopen %s: %s\n", opt.outputfile, strerror(errno));
      exit(EXIT_FAILURE);
    }
  } else {
  /* otherwise use stdout */
    output = stdout;
  }

/* OUTPUT DATA */

  /* output contains file handle for output */

  if (!opt.use_out) 
    fprintf(output, "\n");

  flds = fields;
  while (flds != NULL) {
    fprintf(output, "%s\t", flds->keyname);
    flds = flds->next;
  }
  fprintf(output, "\n");
  show_list(output);

  fprintf(output, "\n");
  fprintf(output, "%lu\n", opt.report_rows);
  fprintf(output, "%lu\n", opt.matched_entries);
  fprintf(output, "%li\n", (long)hostsupdtime);
  fprintf(output, "%li\n", (long)servicesupdtime);
  fprintf(output, "%li\n", (long)populatecachetime);

  /* clean up */

  free_conn_data();
  free_input_file();

  if (opt.use_out) {
    if (opt.verbose >= VERBOSE_WARNING)
      fprintf(stderr, "Closing '%s'\n", opt.outputfile);

    retval = fclose(output);
    if (retval == EOF) {
      perror("fclose");
    }
  }
}
