/* $Id: reportdef.c 643 2016-08-03 22:40:45Z bhockney $ */

/* (C) 2004-2016 by Bob Hockney <zeus@ix.netcom.com>    *
 *                                                      *
 * report definition parser for wfwl_syslog             *
 *                                                      *
 * This code is distributed under the terms of GNU GPL  */

#include <errno.h>
#include <sys/stat.h>
#include <string.h>
#include <ctype.h>
#include <netdb.h>
#include <math.h>

#include <reportdef.h>
#include <main.h>
#include <utils.h>
#include <getdate.h>

struct sort_order *sort = NULL;
struct field_order *fields = NULL;
struct selection *selection = NULL;
struct fields_used fields_used;
struct select_sum select_sum;
extern struct options opt;

unsigned int sort_count = 0;
unsigned int column_count = 0;
unsigned int selection_count = 0;

/* This array controls how report definition keys are parsed for sorting,       *
 * criteria, and columns. Fields in the array are                               *
 *         key:             Key string                                          *
 *         sort_key:        Sort index                                          *
 *         sel_key:         Selection index                                     *
 *         sel_type:        Selection type                                      *
 *         aggregate_value: Pointer for selection aggregate value               *
 *         field_used:      Pointer to fields_used struct for columns           *
 *         col_key:         Column index                                        *
 *         cache_resolve:   Cache constant for host and service columns         *
 *         use_geoip:       whether source, dest or neither geoip fields needed *
 *                                                                              */
struct reportdef_keys keys [] = {
  {"ahesp_spi",SORT_AHESP_SPI,SEL_AHESP_SPI,SEL_NUM,NULL,&fields_used.ahesp_spi,COL_AHESP_SPI,0,0},
  {"icmptype",SORT_ICMP_TYPE,0,0,NULL,&fields_used.icmp_type,COL_ICMP_TYPE,0,0},
  {"icmpcode",SORT_ICMP_CODE,0,0,NULL,&fields_used.icmp_code,COL_ICMP_CODE,0,0},
  {"icmpechoid",SORT_ICMP_ECHOID,0,0,NULL,&fields_used.icmp_echoid,COL_ICMP_ECHOID,0,0},
  {"icmpechoseq",SORT_ICMP_ECHOSEQ,0,0,NULL,&fields_used.icmp_echoseq,COL_ICMP_ECHOSEQ,0,0},
  {"icmp_code",0,SEL_ICMP_CODE,SEL_NUM,NULL,NULL,0,0,0},
  {"icmp_type",0,SEL_ICMP_TYPE,SEL_NUM,NULL,NULL,0,0,0},
  {"icmp_echoid",0,SEL_ICMP_ECHOID,SEL_NUM,NULL,NULL,0,0,0},
  {"icmp_echoseq",0,SEL_ICMP_ECHOSEQ,SEL_NUM,NULL,NULL,0,0,0},
  {"icmpv6_code",0,SEL_ICMPv6_CODE,SEL_NUM,NULL,NULL,0,0,0},
  {"icmpv6_type",0,SEL_ICMPv6_TYPE,SEL_NUM,NULL,NULL,0,0,0},
  {"icmpv6_echoid",0,SEL_ICMPv6_ECHOID,SEL_NUM,NULL,NULL,0,0,0},
  {"icmpv6_echoseq",0,SEL_ICMPv6_ECHOSEQ,SEL_NUM,NULL,NULL,0,0,0},
  {"icmp_fragmtu",SORT_ICMP_MTU,SEL_ICMP_FRAGMTU,SEL_NUM,NULL,&fields_used.icmp_mtu,COL_ICMP_MTU,0,0},
  {"icmp_gateway",SORT_ICMP_GW,SEL_ICMP_GATEWAY,SEL_IPADDR,NULL,&fields_used.icmp_gw,COL_ICMP_GATEWAY,0,0},
  {"ip_csum",SORT_CSUM,SEL_IP_CSUM,SEL_NUM,NULL,&fields_used.csum,COL_IP_CSUM,0,0},
  {"ip_fragoff",SORT_FRAGOFF,SEL_IP_FRAGOFF,SEL_NUM,NULL,&fields_used.fragoff,COL_IP_FRAGOFF,0,0},
  {"ip_id",SORT_IP_ID,SEL_IP_ID,SEL_NUM,NULL,&fields_used.ipid,COL_IP_ID,0,0},
  {"ip_ihl",SORT_IP_IHL,SEL_IP_IHL,SEL_NUM,NULL,&fields_used.ihl,COL_IP_IHL,0,0},
  {"ip_protocol",SORT_PROTOCOL,SEL_PROTOCOL,SEL_PROTO,NULL,&fields_used.protocol,COL_IP_PROTO,0,0},
  {"ip_daddr",SORT_DESTHOST,SEL_DESTHOST,SEL_IPADDR,NULL,&fields_used.dhost,COL_DESTHOST,0,0},
  {"ip_saddr",SORT_SOURCEHOST,SEL_SOURCEHOST,SEL_IPADDR,NULL,&fields_used.shost,COL_SOURCEHOST,0,0},
  {"ip_tos",SORT_IP_TOS,SEL_IP_TOS,SEL_NUM,NULL,&fields_used.tos,COL_IP_TOS,0,0},
  {"ip_totlen",SORT_IP_TOTLEN,SEL_IP_TOTLEN,SEL_NUM,NULL,&fields_used.totlen,COL_IP_LEN,0,0},
  {"ip_ttl",SORT_IP_TTL,SEL_IP_TTL,SEL_NUM,NULL,&fields_used.ttl,COL_IP_TTL,0,0},
  {"ip_df",0,SEL_IP_DF,SEL_DFMF,NULL,NULL,0,0,0},
  {"ip_mf",0,SEL_IP_MF,SEL_DFMF,NULL,NULL,0,0,0},
  {"ip6_payloadlen",SORT_IP6_PAY_LEN,SEL_IP6_PAY_LEN,SEL_NUM,NULL,&fields_used.pay_len,COL_IP6_PAY_LEN,0,0},
  {"ip6_flowlabel",SORT_IP6_FLOWLABEL,SEL_IP6_FLOWLABEL,SEL_NUM,NULL,&fields_used.flowlabel,COL_IP6_FLOWLABEL,0,0},
  {"dport",SORT_DESTPORT,0,0,NULL,&fields_used.dport,COL_DESTPORT,0,0},
  {"sport",SORT_SOURCEPORT,0,0,NULL,&fields_used.sport,COL_SOURCEPORT,0,0},
  {"tcp_dport",0,SEL_TCP_DPORT,SEL_NUM,NULL,NULL,0,0,0},
  {"tcp_sport",0,SEL_TCP_SPORT,SEL_NUM,NULL,NULL,0,0,0},
  {"udp_dport",0,SEL_UDP_DPORT,SEL_NUM,NULL,NULL,0,0,0},
  {"udp_sport",0,SEL_UDP_SPORT,SEL_NUM,NULL,NULL,0,0,0},
  {"dst_host",0,0,0,NULL,&fields_used.dhost_name,COL_DESTHOST_NAME,CACHE_RESOLVE,0},
  {"src_host",0,0,0,NULL,&fields_used.shost_name,COL_SOURCEHOST_NAME,CACHE_RESOLVE,0},
  {"dst_service",0,0,0,NULL,&fields_used.dst_service,COL_DEST_SERVICE,CACHE_RESOLVE,0},
  {"src_service",0,0,0,NULL,&fields_used.src_service,COL_SOURCE_SERVICE,CACHE_RESOLVE,0},
  {"tcp_ackseq",SORT_TCP_ACKSEQ,SEL_TCP_ACKSEQ,SEL_NUM,NULL,&fields_used.tcp_ack_seq,COL_TCP_ACKSEQ,0,0},
  {"tcp_options",SORT_TCP_OPTS,SEL_TCP_OPTS_EXACT,SEL_BOOLEAN,NULL,&fields_used.flags,COL_TCP_OPTIONS,0,0},
  {"tcp_seq",SORT_TCP_SEQ,SEL_TCP_SEQ,SEL_NUM,NULL,&fields_used.tcp_seq,COL_TCP_SEQ,0,0},
  {"tcp_urgp",SORT_TCP_URGP,SEL_TCP_URGP,SEL_NUM,NULL,&fields_used.tcp_urgp,COL_TCP_URGP,0,0},
  {"tcp_window",SORT_TCP_WINDOW,SEL_TCP_WINDOW,SEL_NUM,NULL,&fields_used.tcp_window,COL_TCP_WINDOW,0,0},
  {"tcp_ack",0,SEL_TCP_ACK,SEL_BOOLEAN,NULL,NULL,0,0,0},
  {"tcp_fin",0,SEL_TCP_FIN,SEL_BOOLEAN,NULL,NULL,0,0,0},
  {"tcp_psh",0,SEL_TCP_PSH,SEL_BOOLEAN,NULL,NULL,0,0,0},
  {"tcp_rst",0,SEL_TCP_RST,SEL_BOOLEAN,NULL,NULL,0,0,0},
  {"tcp_syn",0,SEL_TCP_SYN,SEL_BOOLEAN,NULL,NULL,0,0,0},
  {"tcp_urg",0,SEL_TCP_URG,SEL_BOOLEAN,NULL,NULL,0,0,0},
  {"tcp_flags",0,SEL_TCP_FLAGS,SEL_BOOLEAN,NULL,NULL,0,0,0},
  {"udp_len",SORT_UDP_LEN,SEL_UDP_LEN,SEL_NUM,NULL,&fields_used.udp_len,COL_UDP_LEN,0,0},
  {"local_hostname",SORT_LOCAL_HOST,SEL_LOCAL_HOST,SEL_REGEX,NULL,&fields_used.hostname,COL_LOCAL_HOST,0,0},
  {"oob_in",SORT_IN_IF,SEL_IN_IF,SEL_REGEX,NULL,&fields_used.inif,COL_IN_IF,0,0},
  {"oob_mark",SORT_FWMARK,SEL_OOB_MARK,SEL_NUM,NULL,&fields_used.fwmark,COL_FWMARK,0,0},
  {"oob_out",SORT_OUT_IF,SEL_OUT_IF,SEL_REGEX,NULL,&fields_used.outif,COL_OUT_IF,0,0},
  {"oob_prefix",SORT_PREFIX,SEL_PREFIX,SEL_REGEX,NULL,&fields_used.log_label,COL_LOG_PREFIX,0,0},
  {"oob_family",SORT_OOB_FAMILY,SEL_OOB_FAMILY,SEL_NUM,NULL,&fields_used.family,COL_OOB_FAMILY,0,0},
  {"oob_protocol",SORT_OOB_PROTOCOL,SEL_OOB_PROTOCOL,SEL_NUM,NULL,&fields_used.oob_protocol,COL_OOB_PROTOCOL,0,0},
  {"mac_saddr_str",SORT_MAC_SADDR,SEL_MAC_DADDR,SEL_REGEX,NULL,&fields_used.mac_saddr_str,COL_MAC_SADDR,0,0},
  {"mac_daddr_str",SORT_MAC_DADDR,SEL_MAC_SADDR,SEL_REGEX,NULL,&fields_used.mac_daddr_str,COL_MAC_DADDR,0,0},
  {"raw_mac",SORT_MAC,SEL_MAC,SEL_REGEX,NULL,&fields_used.raw_mac,COL_MAC,0,0},
  {"id",SORT_ID,0,SEL_ID,NULL,NULL,COL_ID,0,0},
  {"sname",SORT_SNAME,0,0,NULL,&fields_used.sname,COL_SNAME,0,0},
  {"count",SORT_COUNT,0,0,NULL,&fields_used.count,COL_COUNT,0,0},
  {"min_count",0,0,SEL_AGG_COUNT,(time_t *)&select_sum.min_count,NULL,0,0,0},
  {"max_count",0,0,SEL_AGG_COUNT,(time_t *)&select_sum.max_count,NULL,0,0,0},
  {"local_time",SORT_LOCAL_TIME,0,0,NULL,&fields_used.local_time,COL_LOCAL_TIME,0,0},
  {"earliest",SORT_START_TIME,0,0,NULL,&fields_used.earliest,COL_START_TIME,0,0},
  {"latest",SORT_END_TIME,0,0,NULL,&fields_used.latest,COL_END_TIME,0,0},
  {"min_date",0,SEL_MIN_DATE,SEL_DATE,NULL,NULL,0,0,0},
  {"max_date",0,SEL_MAX_DATE,SEL_DATE,NULL,NULL,0,0,0},
  {"min_latest",0,0,SEL_AGG_DATE,&select_sum.min_latest,NULL,0,0,0},
  {"max_earliest",0,0,SEL_AGG_DATE,&select_sum.max_earliest,NULL,0,0,0},
  {"continent_code_s",SORT_CONTINENT_CODE_SRC,SEL_CONTINENT_CODE_SRC,SEL_REGEX,NULL,&fields_used.continent_code_s,COL_CONTINENT_CODE_SRC,0,GEOIP_USE_SRC},
  {"continent_name_s",SORT_CONTINENT_NAME_SRC,SEL_CONTINENT_NAME_SRC,SEL_REGEX,NULL,&fields_used.continent_name_s,COL_CONTINENT_NAME_SRC,0,GEOIP_USE_SRC},
  {"country_iso_code_s",SORT_COUNTRY_ISO_CODE_SRC,SEL_COUNTRY_ISO_CODE_SRC,SEL_REGEX,NULL,&fields_used.country_iso_code_s,COL_COUNTRY_ISO_CODE_SRC,0,GEOIP_USE_SRC},
  {"country_name_s",SORT_COUNTRY_NAME_SRC,SEL_COUNTRY_NAME_SRC,SEL_REGEX,NULL,&fields_used.country_name_s,COL_COUNTRY_NAME_SRC,0,GEOIP_USE_SRC},
  {"subdivision_1_iso_code_s",SORT_SUBDIVISION_1_ISO_CODE_SRC,SEL_SUBDIVISION_1_ISO_CODE_SRC,SEL_REGEX,NULL,&fields_used.subdivision_1_iso_code_s,COL_SUBDIVISION_1_ISO_CODE_SRC,0,GEOIP_USE_SRC},
  {"subdivision_1_name_s",SORT_SUBDIVISION_1_NAME_SRC,SEL_SUBDIVISION_1_NAME_SRC,SEL_REGEX,NULL,&fields_used.subdivision_1_name_s,COL_SUBDIVISION_1_NAME_SRC,0,GEOIP_USE_SRC},
  {"subdivision_2_iso_code_s",SORT_SUBDIVISION_2_ISO_CODE_SRC,SEL_SUBDIVISION_2_ISO_CODE_SRC,SEL_REGEX,NULL,&fields_used.subdivision_2_iso_code_s,COL_SUBDIVISION_2_ISO_CODE_SRC,0,GEOIP_USE_SRC},
  {"subdivision_2_name_s",SORT_SUBDIVISION_2_NAME_SRC,SEL_SUBDIVISION_2_NAME_SRC,SEL_REGEX,NULL,&fields_used.subdivision_2_name_s,COL_SUBDIVISION_2_NAME_SRC,0,GEOIP_USE_SRC},
  {"city_name_s",SORT_CITY_NAME_SRC,SEL_CITY_NAME_SRC,SEL_REGEX,NULL,&fields_used.city_name_s,COL_CITY_NAME_SRC,0,GEOIP_USE_SRC},
  {"metro_code_s",SORT_METRO_CODE_SRC,SEL_METRO_CODE_SRC,SEL_NUM,NULL,&fields_used.metro_code_s,COL_METRO_CODE_SRC,0,GEOIP_USE_SRC},
  {"accuracy_radius_s",SORT_ACCURACY_RADIUS_SRC,SEL_ACCURACY_RADIUS_SRC,SEL_NUM,NULL,&fields_used.accuracy_radius_s,COL_ACCURACY_RADIUS_SRC,0,GEOIP_USE_SRC},
  {"time_zone_s",SORT_TIME_ZONE_SRC,SEL_TIME_ZONE_SRC,SEL_REGEX,NULL,&fields_used.time_zone_s,COL_TIME_ZONE_SRC,0,GEOIP_USE_SRC},
  {"registered_country_geoname_id_s",SORT_REGISTERED_COUNTRY_GEONAME_ID_SRC,SEL_REGISTERED_COUNTRY_GEONAME_ID_SRC,SEL_REGEX,NULL,&fields_used.registered_country_geoname_id_s,COL_REGISTERED_COUNTRY_GEONAME_ID_SRC,0,GEOIP_USE_SRC},
  {"represented_country_geoname_id_s",SORT_REPRESENTED_COUNTRY_GEONAME_ID_SRC,SEL_REPRESENTED_COUNTRY_GEONAME_ID_SRC,SEL_REGEX,NULL,&fields_used.represented_country_geoname_id_s,COL_REPRESENTED_COUNTRY_GEONAME_ID_SRC,0,GEOIP_USE_SRC},
  {"is_anonymous_proxy_s",SORT_IS_ANONYMOUS_PROXY_SRC,SEL_IS_ANONYMOUS_PROXY_SRC,SEL_BOOLEAN,NULL,&fields_used.is_anonymous_proxy_s,COL_IS_ANONYMOUS_PROXY_SRC,0,GEOIP_USE_SRC},
  {"is_satellite_provider_s",SORT_IS_SATELLITE_PROVIDER_SRC,SEL_IS_SATELLITE_PROVIDER_SRC,SEL_BOOLEAN,NULL,&fields_used.is_satellite_provider_s,COL_IS_SATELLITE_PROVIDER_SRC,0,GEOIP_USE_SRC},
  {"postal_code_s",SORT_POSTAL_CODE_SRC,SEL_POSTAL_CODE_SRC,SEL_REGEX,NULL,&fields_used.postal_code_s,COL_POSTAL_CODE_SRC,0,GEOIP_USE_SRC},
  {"latitude_s",SORT_LATITUDE_SRC,SEL_LATITUDE_SRC,SEL_DOUBLE,NULL,&fields_used.latitude_s,COL_LATITUDE_SRC,0,GEOIP_USE_SRC},
  {"longitude_s",SORT_LONGITUDE_SRC,SEL_LONGITUDE_SRC,SEL_DOUBLE,NULL,&fields_used.longitude_s,COL_LONGITUDE_SRC,0,GEOIP_USE_SRC},
  {"continent_code_d",SORT_CONTINENT_CODE_DST,SEL_CONTINENT_CODE_DST,SEL_REGEX,NULL,&fields_used.continent_code_d,COL_CONTINENT_CODE_DST,0,GEOIP_USE_DST},
  {"continent_name_d",SORT_CONTINENT_NAME_DST,SEL_CONTINENT_NAME_DST,SEL_REGEX,NULL,&fields_used.continent_name_d,COL_CONTINENT_NAME_DST,0,GEOIP_USE_DST},
  {"country_iso_code_d",SORT_COUNTRY_ISO_CODE_DST,SEL_COUNTRY_ISO_CODE_DST,SEL_REGEX,NULL,&fields_used.country_iso_code_d,COL_COUNTRY_ISO_CODE_DST,0,GEOIP_USE_DST},
  {"country_name_d",SORT_COUNTRY_NAME_DST,SEL_COUNTRY_NAME_DST,SEL_REGEX,NULL,&fields_used.country_name_d,COL_COUNTRY_NAME_DST,0,GEOIP_USE_DST},
  {"subdivision_1_iso_code_d",SORT_SUBDIVISION_1_ISO_CODE_DST,SEL_SUBDIVISION_1_ISO_CODE_DST,SEL_REGEX,NULL,&fields_used.subdivision_1_iso_code_d,COL_SUBDIVISION_1_ISO_CODE_DST,0,GEOIP_USE_DST},
  {"subdivision_1_name_d",SORT_SUBDIVISION_1_NAME_DST,SEL_SUBDIVISION_1_NAME_DST,SEL_REGEX,NULL,&fields_used.subdivision_1_name_d,COL_SUBDIVISION_1_NAME_DST,0,GEOIP_USE_DST},
  {"subdivision_2_iso_code_d",SORT_SUBDIVISION_2_ISO_CODE_DST,SEL_SUBDIVISION_2_ISO_CODE_DST,SEL_REGEX,NULL,&fields_used.subdivision_2_iso_code_d,COL_SUBDIVISION_2_ISO_CODE_DST,0,GEOIP_USE_DST},
  {"subdivision_2_name_d",SORT_SUBDIVISION_2_NAME_DST,SEL_SUBDIVISION_2_NAME_DST,SEL_REGEX,NULL,&fields_used.subdivision_2_name_d,COL_SUBDIVISION_2_NAME_DST,0,GEOIP_USE_DST},
  {"city_name_d",SORT_CITY_NAME_DST,SEL_CITY_NAME_DST,SEL_REGEX,NULL,&fields_used.city_name_d,COL_CITY_NAME_DST,0,GEOIP_USE_DST},
  {"metro_code_d",SORT_METRO_CODE_DST,SEL_METRO_CODE_DST,SEL_NUM,NULL,&fields_used.metro_code_d,COL_METRO_CODE_DST,0,GEOIP_USE_DST},
  {"accuracy_radius_d",SORT_ACCURACY_RADIUS_DST,SEL_ACCURACY_RADIUS_DST,SEL_NUM,NULL,&fields_used.accuracy_radius_d,COL_ACCURACY_RADIUS_DST,0,GEOIP_USE_DST},
  {"time_zone_d",SORT_TIME_ZONE_DST,SEL_TIME_ZONE_DST,SEL_REGEX,NULL,&fields_used.time_zone_d,COL_TIME_ZONE_DST,0,GEOIP_USE_DST},
  {"registered_country_geoname_id_d",SORT_REGISTERED_COUNTRY_GEONAME_ID_DST,SEL_REGISTERED_COUNTRY_GEONAME_ID_DST,SEL_REGEX,NULL,&fields_used.registered_country_geoname_id_d,COL_REGISTERED_COUNTRY_GEONAME_ID_DST,0,GEOIP_USE_DST},
  {"represented_country_geoname_id_d",SORT_REPRESENTED_COUNTRY_GEONAME_ID_DST,SEL_REPRESENTED_COUNTRY_GEONAME_ID_DST,SEL_REGEX,NULL,&fields_used.represented_country_geoname_id_d,COL_REPRESENTED_COUNTRY_GEONAME_ID_DST,0,GEOIP_USE_DST},
  {"is_anonymous_proxy_d",SORT_IS_ANONYMOUS_PROXY_DST,SEL_IS_ANONYMOUS_PROXY_DST,SEL_BOOLEAN,NULL,&fields_used.is_anonymous_proxy_d,COL_IS_ANONYMOUS_PROXY_DST,0,GEOIP_USE_DST},
  {"is_satellite_provider_d",SORT_IS_SATELLITE_PROVIDER_DST,SEL_IS_SATELLITE_PROVIDER_DST,SEL_BOOLEAN,NULL,&fields_used.is_satellite_provider_d,COL_IS_SATELLITE_PROVIDER_DST,0,GEOIP_USE_DST},
  {"postal_code_d",SORT_POSTAL_CODE_DST,SEL_POSTAL_CODE_DST,SEL_REGEX,NULL,&fields_used.postal_code_d,COL_POSTAL_CODE_DST,0,GEOIP_USE_DST},
  {"latitude_d",SORT_LATITUDE_DST,SEL_LATITUDE_DST,SEL_DOUBLE,NULL,&fields_used.latitude_d,COL_LATITUDE_DST,0,GEOIP_USE_DST},
  {"longitude_d",SORT_LONGITUDE_DST,SEL_LONGITUDE_DST,SEL_DOUBLE,NULL,&fields_used.longitude_d,COL_LONGITUDE_DST,0,GEOIP_USE_DST},
  {"extra",0,0,0,NULL,NULL,0,0,0},
  {"extra_value",0,0,0,NULL,NULL,0,0,0},
  {"oob_earliest",0,0,0,NULL,NULL,0,0,0},
  {"oob_latest",0,0,0,NULL,NULL,0,0,0},
  {"oob_time_sec",0,0,0,NULL,NULL,0,0,0},
  {"oob_time_usec",0,0,0,NULL,NULL,0,0,0},
  {"oob_max_date",0,0,0,NULL,NULL,0,0,0},
  {"oob_min_date",0,0,0,NULL,NULL,0,0,0},
  {"oob_max_earliest",0,0,0,NULL,NULL,0,0,0},
  {"oob_min_latest",0,0,0,NULL,NULL,0,0,0},
  {NULL,0,0,0,NULL,NULL,0,0,0}
};

/* Function to get string parameter.
 *
 * Leading whitespace is discarded.  If the next character is a
 * single or double quote, then line is taken until last matching
 * quote is encountered, with embedded quotes left intact.
 * Anything trailing the last matching quote is discarded.
 * If no matching quote is found, then the rest of the line is
 * returned including the first quote character, but with trailing
 * whitespace discarded.  Leading and trailing whitespace is 
 * included if within matching quotes.
 */

char * str_parameter(char *string)
{
  char *quote, *startquote = NULL, *endquote, *pnt, *ret;

  ret = string;

  while (*ret == ' ' || *ret == '\t')
    ++ret;

  quote = strdup(ret);
  if ( *ret == '"' || *ret == '\'') {
    startquote = ret;
    xstrncpy(quote, ret, 1 + 1);
  } else {
    *quote = '\0';
    startquote = ret;
  }

  ++ret; 
  endquote = strrchr(ret, *quote);
  pnt = ret;
  while ((endquote == NULL || *endquote == '\0' || endquote != pnt) && *pnt != '\n' && *pnt != '\0')
    ++pnt;

  /* When pnt reaches matching end quote, or end of string, terminator is  *
   * written over current character.  If matching quote was found, opening *
   * quote is overwritten with space.  Trailing whitespace, if any, is     *
   * discarded                                                             */

  *pnt = '\0';

  if ((endquote != NULL && (*quote == '"' || *quote == '\'') && endquote == pnt)) {
    *startquote = ' ';
  } else {
    --ret;
  }

  if (*endquote == '\0' || endquote != pnt) { 
    --pnt;
    while (*pnt == ' ' || *pnt == '\t') {
      *pnt = '\0';
      --pnt;
    }
  }

  free(quote);
  return ret;
}

double double_parameter(char *string, int linenum)
{
  char *pnt, *para; 

  if (string == NULL) return 0;

  para = string;

  while (*para == ' ' || *para == '\t' || *para == '"')
    ++para;

  pnt = para;
  while (*pnt != '\n' && *pnt != ' ' && *pnt != '"' && *pnt != ':'
         && *pnt != '#' && *pnt != ',' && *pnt != '\t' && *pnt != '\0') {
    if(!isxdigit((int)*pnt) && strncasecmp(pnt, "x", 1) != 0 && strncmp(pnt, ".", 1) != 0 && strncmp(pnt, "-", 1) != 0) {
      if (opt.verbose >= VERBOSE_INFO)
        fprintf(stderr, "Error in report definition, line %d: %s not a double\n", linenum, pnt);
        return 0;
      }
  
    ++pnt;
  }
  *pnt = '\0';

  return strtod (para, NULL);
}

unsigned long int num_parameter(char *string, int linenum)
{
  char *pnt, *para, *retptr; 
  unsigned long ret;

  if (string == NULL) return 0;

  para = string;

  while (*para == ' ' || *para == '\t' || *para == '"')
    ++para;

  pnt = para;
  while (*pnt != '\n' && *pnt != ' ' && *pnt != '"' && *pnt != ':'
         && *pnt != '#' && *pnt != ',' && *pnt != '\t' && *pnt != '\0') {
    if(!isxdigit((int)*pnt) && strncasecmp(pnt, "x", 1) != 0 && strncasecmp(pnt, "b", 1) != 0) {
      if (opt.verbose >= VERBOSE_INFO)
        fprintf(stderr, "Error in report definition, line %d: %s not a number\n", linenum, pnt);
        return 0;
      }
  
    ++pnt;
  }
  *pnt = '\0';

  ret = strtoul (para, &retptr, 0);

  if (*retptr == '\0')
    return ret;
  else if (strncasecmp(retptr, "b", 1) == 0)
    ret = strtoul (++retptr, NULL, 2);
  else
    ret = 0;

  return ret;
}

unsigned char yes_or_no(char *string, int linenum)
{
  char *pnt, *orig;

  orig = strdup(string);

  while (*string == ' ' || *string == '\t' || *string == '"')
    ++string;

  pnt = string;

  while (*pnt != '\n' && *pnt != ' ' && *pnt != '"' && *pnt != '#' && *pnt != '\t' && *pnt != '\0')
    ++pnt;
  *pnt = '\0';

  if((strncasecmp(string, "yes", 3) == 0)
     || (strncasecmp(string, "on", 2) == 0)
     || (strncasecmp(string, "true", 4) == 0)) {
    if (opt.verbose >= VERBOSE_DEBUG)
      fprintf(stderr, "%s was parsed to %s and evaluated to true\n", orig, string);

    free(orig);
    return true;
  } else if((strlen(string) == 0)
            || (strncasecmp(string, "no", 2) == 0)
            || (strncasecmp(string, "off", 3) == 0)
            || (strncasecmp(string, "false", 5) == 0)) {
    if (opt.verbose >= VERBOSE_DEBUG)
      fprintf(stderr, "%s was parsed to %s and evaluated to false\n", orig, string);

    free(orig);
    return false;
  } else {
    if (opt.verbose >= VERBOSE_DEBUG)
      fprintf(stderr, "Error in report definition, line %d, assuming 'true'\n", linenum);

    free(orig);
    return true;
  }
}

int resolve_protobyname(char *proto)
{
  struct protoent *protoent;

  proto = str_parameter(proto);
  protoent = getprotobyname(proto);
  if (protoent != NULL) {
    return (protoent->p_proto);
  } else {
    return 0;
  }
}

/* Adds an item to the fields list. */
struct field_order * add_column(struct field_order *new, char *key, char *parm, int linenum)
{
  struct field_order *col, *list = NULL, *last = NULL, *this = NULL, *next = NULL;
  int j;

  for (j=0; keys[j].key != NULL; j++) {
    if(strcmp(keys[j].key, key) == 0)
      break;
  }

  list = new;

  if (keys[j].key == NULL) {
    if (opt.verbose >= VERBOSE_ALERT)
      fprintf(stderr, "Skipping unrecognized column: %s\n", strip_nl(key));

    return list;
  }

  if (!keys[j].col_key) {
    if (opt.verbose >= VERBOSE_ERROR)
      fprintf(stderr, "ignoring key = %s\n", key);

    return list;
  }

  col = xmalloc(sizeof(struct field_order));
  col->next = NULL;
  xstrncpy(col->keyname, keys[j].key, SHORTLEN);
  col->position = num_parameter(parm, linenum);
  if (keys[j].field_used != NULL) *keys[j].field_used = 1;
  col->field = keys[j].col_key;
  if (keys[j].cache_resolve)
    opt.resolve_hosts |= keys[j].cache_resolve;
  opt.use_geoip |= keys[j].use_geoip;

  column_count++;
  last = list;
  this = list;
  next = NULL;
  if (this != NULL) next = this->next;

  while (this != NULL) {
    if (opt.verbose >= VERBOSE_ERROR) {
    fprintf(stderr, "Inserting column %s into list: this->position = %i\n", key, this->position);
    }
    if (col->position > this->position) {
    last = this;
    this = last->next;
    if (this != NULL) next = this->next;
    continue;
    } else {
    break;
    }
  }

  if (opt.verbose >= VERBOSE_DEBUG) {
    if (last != NULL) fprintf(stderr, "Last->position: %i\n", last->position);
    if (this != NULL) fprintf(stderr, "This->position: %i\n", this->position);
    if (next != NULL) fprintf(stderr, "Next->position: %i\n", next->position);
  }

  if (list == NULL) {
    if (opt.verbose >= VERBOSE_DEBUG) {
    fprintf(stderr, "First item seen; position: %i\n", col->position);
    }
    list = col; 
  } else { /* not first item seen */
    if (list->next == NULL) {
    if (opt.verbose >= VERBOSE_DEBUG) {
      fprintf(stderr, "Second item seen; position: %i\n", col->position);
    }
    if (col->position > list->position) {
      list->next = col;
    } else {
      this = list;
      list = col;
      list->next = this;
    }
    } else { /* not second item seen */
    if (opt.verbose >= VERBOSE_DEBUG) {
      fprintf(stderr, "Not first or second item; position: %i\n", col->position);
    }
    if (this == NULL) {
      if (opt.verbose >= VERBOSE_DEBUG) {
      fprintf(stderr, "this->position = NULL = place item last\n");
      }
      last->next = col;
      col->next = NULL;
    } else { /* not last */
      if (this->position == last->position) {
      if (opt.verbose >= VERBOSE_DEBUG) {
        fprintf(stderr, "this->position = last->position = place item first\n");
      }
      col->next = list;
      list = col;
      } else { /* not first */
      if (opt.verbose >= VERBOSE_DEBUG) {
        fprintf(stderr, "this->position > last->position = insert item\n");
      }
      last->next = col;
      col->next = this;
      } /* end: else clause of test to place item first */
    } /* end: else clause of test to place item last */
    } /* end: else clause of test for second item seen */
  } /* end: else clause of test for first item seen */

  if (opt.verbose >= VERBOSE_WARNING)
    fprintf(stderr, "Added column %s at position %i\n", key, col->position);

  return list;
}

/* Adds an item to the criteria list.                       *
 * If selection->have_value = -1, then only invert flag was *
 *   present and entry should not be used for selection     */
void add_selection(char *key, char *parm, int linenum, int mode)
{
  struct selection *sel = NULL, *addl = NULL, *prior = NULL;
  int seen, bits, res, skip = 0;
  static int fragoff_flags = 0;
  char *ptr = NULL, *ptrr, *parms;
  struct in_addr inp;
#if HAVE_INET_NTOP
  struct in6_addr inp6;
#endif
  char buf[INET6_ADDRSTRLEN], re_err[SHORTLEN], delim;
  char addr[16], mask[16];
  int j;

  *buf = '\0';

  for (j=0; keys[j].key != NULL; j++) {
    if(strcmp(keys[j].key, key) == 0)
      break;
  }

  /* skip any unrecognized criterion */
  if (keys[j].key == NULL) {
    if (opt.verbose >= VERBOSE_ALERT)
      fprintf(stderr, "Skipping unrecognized criterion: %s\n", strip_nl(key));

    return;
  }

  if (!keys[j].sel_type) {
    if (opt.verbose >= VERBOSE_ERROR)
      fprintf(stderr, "ignoring key = %s\n", key);

    return;
  }

  sel = xmalloc(sizeof(struct selection));
  sel->value = 0;
  sel->max_value = 0;
  sel->double_value = 0;
  sel->max_double_value = 0;
  *sel->svalue = '\0';
  sel->in_addr = 0;
  sel->netmask = -1;
#if HAVE_INET_NTOP
  memset(sel->in6_addr.s6_addr, 0x00, sizeof(struct in6_addr));
#endif
  sel->in6_bits = 128;
  sel->family = AF_INET;
  regcomp(&sel->buf, ".", REG_NOSUB);
  sel->invert = 0;
  sel->field = -1;
  sel->type = 0;
  sel->next = NULL;
  xstrncpy(sel->keyname, key, SHORTLEN);
  parms = strdup(parm);

  switch(keys[j].sel_type) {
    case SEL_DATE:
    case SEL_BOOLEAN:
    case SEL_DFMF:
      if (mode != DEF_WHERE) {
        if (opt.verbose >= VERBOSE_ALERT)
          fprintf(stderr, "Skipping invert flag for criterion: %s\n", key);

        free(sel);
        return;
      }
      /* fall through */
    case SEL_NUM:
    case SEL_DOUBLE:
    case SEL_IPADDR:
    case SEL_REGEX:
    case SEL_PROTO:
      sel->field = keys[j].sel_key;
      if (keys[j].sel_type == SEL_DFMF)
        sel->type = SEL_NUM;
      else
        sel->type = keys[j].sel_type;
      break;
    case SEL_AGG_COUNT:
      *keys[j].aggregate_value = num_parameter(parm, linenum);
      free(sel);
      return;
    case SEL_AGG_DATE:
      *keys[j].aggregate_value = get_date(str_parameter(parm), NULL);
      free(sel);
      return;
    case SEL_ID:
      opt.packet = num_parameter(parm, linenum);
      free(sel);
      return;
    default:
      break;
  }

  switch (mode) {
    case DEF_INVERT:
      sel->invert = yes_or_no(parm, linenum);
      break;
    case DEF_WHERE:
      switch (sel->type) {
        case SEL_PROTO:
          xstrncpy(sel->svalue, str_parameter(parm), SHORTLEN);
          ptr = strpbrk((char *)&sel->svalue, ",");
          if (ptr != NULL)
            *ptr = '\0';
          if(isdigit((unsigned char)*sel->svalue)) {
            sel->value = num_parameter(parm, linenum);
            ptrr = strpbrk((char *)&sel->svalue, ":");
            if (ptrr == NULL) {
              sel->max_value = sel->value;
            } else {
              sel->max_value = num_parameter(++ptrr, linenum);
            }
          } else {
            sel->value = resolve_protobyname(sel->svalue);
            sel->max_value = sel->value;
          }
          if (ptr != NULL)
            parm = ++ptr;
          break;
        case SEL_NUM:
          xstrncpy((char *)&sel->svalue, str_parameter(parm), SHORTLEN);
          sel->value = num_parameter(parm, linenum);
          ptr = strpbrk((char *)&sel->svalue, ",");
          if (ptr != NULL)
            *ptr = '\0';
          ptrr = strpbrk((char *)&sel->svalue, ":");
          if (ptrr == NULL) {
            sel->max_value = sel->value;
          } else {
            sel->max_value = num_parameter(++ptrr, linenum);
          }
          if (ptr != NULL)
            parm = ++ptr;
        break;
        case SEL_DOUBLE:
          xstrncpy((char *)&sel->svalue, str_parameter(parm), SHORTLEN);
          sel->double_value = double_parameter(parm, linenum);
          ptr = strpbrk((char *)&sel->svalue, ",");
          if (ptr != NULL)
            *ptr = '\0';
          ptrr = strpbrk((char *)&sel->svalue, ":");
          if (ptrr == NULL) {
            sel->max_double_value = sel->double_value;
          } else {
            sel->max_double_value = double_parameter(++ptrr, linenum);
          }
          if (ptr != NULL)
            parm = ++ptr;
        break;
        case SEL_IPADDR:
          xstrncpy(sel->svalue, str_parameter(parm), SHORTLEN);
          ptr = strpbrk((char *)&sel->svalue, ",");
          if (ptr != NULL)
            *ptr = '\0';
#if HAVE_INET_NTOP
          if (strpbrk((char *)&sel->svalue, ":"))
            sel->family = AF_INET6;
#endif
          ptrr = strpbrk((char *)&sel->svalue, "/");
          if (ptrr != NULL) {
            strncpy(&delim, ptrr, 1);
            *ptrr = '\0';
          }
          if (sel->family == AF_INET) {
            if (convert_ip(AF_INET, sel->svalue, &inp, NULL, NULL)) { /* error */
              if (opt.verbose >= VERBOSE_NOTICE)
                fprintf(stderr, "Couldn't convert %s\n", sel->svalue);
              break;
            }
            sel->in_addr = ntohl(inp.s_addr);
#if HAVE_INET_NTOP
          } else {
            if (convert_ip(AF_INET6, sel->svalue, NULL, &inp6, NULL)) { /* error */
              if (opt.verbose >= VERBOSE_NOTICE)
                fprintf(stderr, "Couldn't convert %s\n", sel->svalue);
              break;
            }
            memcpy(sel->in6_addr.s6_addr, inp6.s6_addr, sizeof(struct in6_addr));
#endif
          }
          if (ptrr != NULL) {
            strncpy(ptrr, &delim, 1);
            if (sel->family == AF_INET) {
              if (strpbrk(++ptrr, ".") == NULL) {
                sel->netmask = ~(uint32_t)(pow(2, 32 - num_parameter(ptrr, linenum)) - 1);
              } else {
                sel->netmask =  inet_network(str_parameter(ptrr));
              }
              sel->in_addr = sel->in_addr & sel->netmask;
            } else {
              bits = atoi(++ptrr);
              if (bits >= 0 && bits <= 128)
                sel->in6_bits = bits;
            }
          }
          if (ptr != NULL)
            parm = ++ptr;
          break;
        case SEL_REGEX:
          xstrncpy(sel->svalue, str_parameter(parm), SHORTLEN);
          res = regcomp(&sel->buf, sel->svalue, REG_NOSUB||REG_EXTENDED);
          if (res) { /* error */
            if (opt.verbose >= VERBOSE_NOTICE) {
              regerror(res, &sel->buf, re_err, SHORTLEN);
              fprintf(stderr, "Error in regular expression \"%s\" for key %s, skipping:\n  %s\n", sel->svalue, key, re_err);
            }
            regcomp(&sel->buf, ".", REG_NOSUB);
            skip++;
          }
          break;
        case SEL_DATE:
          xstrncpy(sel->svalue, str_parameter(parm), SHORTLEN);
          break;
        case SEL_BOOLEAN:
          sel->value = yes_or_no(parm, linenum);
          break;
        default:
          /* error */
          break;
      } /* end switch (sel->type) */
      opt.use_geoip |= keys[j].use_geoip;
      break;
    default:
      /* error */
      break;
  } /* end switch (mode) */


  if (sel->field == SEL_TCP_FLAGS) {
    sel->value = 0;
    strcpy(sel->keyname, "tcp_flags");
  }

  if (sel->field == SEL_TCP_SYN) {
    sel->value = TCP_SYN;
    strcpy(sel->keyname, "tcp_flags");
    sel->field = SEL_TCP_FLAGS; 
  }

  if (sel->field == SEL_TCP_ACK) {
    sel->value = TCP_ACK;
    strcpy(sel->keyname, "tcp_flags");
    sel->field = SEL_TCP_FLAGS; 
  }

  if (sel->field == SEL_TCP_FIN) {
    sel->value = TCP_FIN;
    strcpy(sel->keyname, "tcp_flags");
    sel->field = SEL_TCP_FLAGS; 
  }

  if (sel->field == SEL_TCP_RST) {
    sel->value = TCP_RST;
    strcpy(sel->keyname, "tcp_flags");
    sel->field = SEL_TCP_FLAGS; 
  }

  if (sel->field == SEL_TCP_URG) {
    sel->value = TCP_URG;
    strcpy(sel->keyname, "tcp_flags");
    sel->field = SEL_TCP_FLAGS; 
  }

  if (sel->field == SEL_TCP_PSH) {
    sel->value = TCP_PSH;
    strcpy(sel->keyname, "tcp_flags");
    sel->field = SEL_TCP_FLAGS; 
  }

  if (sel->field == SEL_TCP_OPTS_EXACT) {
    sel->value = TCP_OPTS_EXACT;
    strcpy(sel->keyname, "tcp_flags");
    sel->field = SEL_TCP_FLAGS; 
  }

  if (sel->field == SEL_IP_DF || sel->field == SEL_IP_MF) {
    fragoff_flags |= sel->value;
    sel->field = SEL_IP_FRAGOFF;
    strcpy(sel->keyname, "ip_fragoff");
  }

  prior = selection;
  seen = 0;
  while (prior != NULL) {
    if (prior->field == sel->field) {
      seen = 1;
      if (mode == DEF_INVERT) {
        prior->invert = sel->invert;
      }

      if (sel->field == SEL_TCP_FLAGS)
        sel->value = sel->value | prior->value;

      if (sel->field == SEL_IP_FRAGOFF) {
        sel->value |= prior->value;
        sel->max_value |= prior->max_value;
      }

      if (mode == DEF_WHERE) {
        prior->have_value = 0;
        prior->value = sel->value; 
        prior->max_value = sel->max_value;
        xstrncpy(prior->svalue, sel->svalue, SHORTLEN);
        sel->invert = prior->invert; /* for multiple add */
        break;
      }
    }
    /* not seen */
    prior = prior->next;
  }

  if (!seen && !skip) {
    selection_count++;
  }

  if (!skip &&(!seen && mode == DEF_INVERT)) { 
    sel->have_value = -1;
  } else {
    sel->have_value = 0;
  }

  if (!seen && !skip) {
    sel->next = selection;
    selection= sel;
  }

  if (opt.verbose >= VERBOSE_WARNING && !skip) {
    switch (mode) {
      case DEF_INVERT:
        fprintf(stderr, "Added invert flag for criterion: %s\n", key);
        break;
      case DEF_WHERE:
        if (sel->field == SEL_TCP_FLAGS) {
          fprintf(stderr, "Added criterion: %s\n", key);
        } else {
          inp.s_addr = htonl(sel->in_addr);
          xstrncpy(addr, inet_ntoa(inp), SHORTLEN);
          inp.s_addr = htonl(sel->netmask);
          xstrncpy(mask, inet_ntoa(inp), SHORTLEN);
          fprintf(stderr, "Added criterion: %s; Value: %lu; Max value: %lu; Svalue: %s; Family: %i; Addr: %s; Netmask: %s;"
#if HAVE_INET_NTOP
            " In6_addr: %s; In6_bits: %u"
#endif
            "\n",
              key, sel->value, sel->max_value, sel->svalue, sel->family, addr, mask
#if HAVE_INET_NTOP
              ,inet_ntop(AF_INET6, sel->in6_addr.s6_addr, buf, INET6_ADDRSTRLEN), sel->in6_bits
#endif
          );
        }
        break;
      default:
        break;
    }
  }

  /* If multiple values are specified, add new selection criterion for each */
  str_parameter(parms); /* drop quotes */
  if (mode == DEF_WHERE && (sel->type == SEL_NUM || sel->type == SEL_DOUBLE || sel->type == SEL_IPADDR || sel->type == SEL_PROTO) && ptr != NULL ) {

    while (ptr != NULL) {

      addl = xmalloc(sizeof(struct selection));
      addl->field = sel->field;
      xstrncpy(addl->keyname, sel->keyname, SHORTLEN);
      addl->type = sel->type;
#if HAVE_INET_NTOP
      memset(addl->in6_addr.s6_addr, 0x00, sizeof(struct in6_addr));
#endif
      addl->in6_bits = 128;
      addl->family = AF_INET;
      regcomp(&addl->buf, ".", REG_NOSUB);
      addl->invert = sel->invert;
      addl->have_value = sel->have_value;
      addl->in_addr = 0;
      addl->netmask = -1;

      xstrncpy((char *)&addl->svalue, str_parameter(parm), SHORTLEN);
      ptr = strpbrk((char *)&addl->svalue, ",");
      if (ptr != NULL)
        *ptr = '\0';
      if (sel-> type == SEL_NUM) {
        addl->value = num_parameter(parm, linenum);
        ptrr = strpbrk((char *)&addl->svalue, ":");
        if (ptrr == NULL) {
          addl->max_value = addl->value;
        } else {
          addl->max_value = num_parameter(++ptrr, linenum);
        }
        if (sel->field == SEL_IP_FRAGOFF) {
          addl->value |= fragoff_flags;
          addl->max_value |= fragoff_flags;
        }
      } else if (sel-> type == SEL_DOUBLE) {
        addl->double_value = double_parameter(parm, linenum);
        ptrr = strpbrk((char *)&addl->svalue, ":");
        if (ptrr == NULL) {
          addl->max_double_value = addl->double_value;
        } else {
          addl->max_double_value = double_parameter(++ptrr, linenum);
        }
      } else if (sel->type == SEL_IPADDR) {
        if (strpbrk((char *)&addl->svalue, ":") == NULL) {
          addl->family = AF_INET;
#if HAVE_INET_NTOP
        } else {
          addl->family = AF_INET6;
#endif
        }
        ptrr = strpbrk((char *)&addl->svalue, "/");
        if (ptrr != NULL) {
          strncpy(&delim, ptrr, 1);
          *ptrr = '\0';
        }
        if (addl->family == AF_INET) {
          if (convert_ip(AF_INET, addl->svalue, &inp, NULL, NULL)) { /* error */
            if (opt.verbose >= VERBOSE_NOTICE)
              fprintf(stderr, "Couldn't convert %s\n", addl->svalue);
            break;
          }
          addl->in_addr = ntohl(inp.s_addr);
#if HAVE_INET_NTOP
        } else {
          if (convert_ip(AF_INET6, addl->svalue, NULL, &inp6, NULL)) { /* error */
            if (opt.verbose >= VERBOSE_NOTICE)
              fprintf(stderr, "Couldn't convert %s\n", addl->svalue);
            break;
          }
          memcpy(addl->in6_addr.s6_addr, inp6.s6_addr, sizeof(struct in6_addr));
#endif
        }
        if (ptrr != NULL) {
          strncpy(ptrr, &delim, 1);
          if (addl->family == AF_INET) {
            if (strpbrk(++ptrr, ".") == NULL) {
              addl->netmask = ~(uint32_t)(pow(2, 32 - num_parameter(ptrr, linenum)) - 1);
            } else {
              addl->netmask = inet_network(str_parameter(ptrr));
            }
            addl->in_addr = addl->in_addr & addl->netmask;
          } else {
            bits = atoi(++ptrr);
            if (bits >= 0 && bits <= 128)
              addl->in6_bits = bits;
          } 
        }
      } else if (sel->type == SEL_PROTO) {
        if(isdigit((unsigned char)*addl->svalue)) {
          addl->value = num_parameter(parm, linenum);
          ptrr = strpbrk((char *)&addl->svalue, ":");
          if (ptrr == NULL) {
            addl->max_value = addl->value;
          } else {
            addl->max_value = num_parameter(++ptrr, linenum);
          }
        } else {
          addl->value = resolve_protobyname(addl->svalue);
          addl->max_value = addl->value;
        }
      }

      if (ptr != NULL)
        parm = ++ptr;

      if (opt.verbose >= VERBOSE_WARNING) {
        inp.s_addr = htonl(addl->in_addr);
        xstrncpy(addr, inet_ntoa(inp), SHORTLEN);
        inp.s_addr = htonl(addl->netmask);
        xstrncpy(mask, inet_ntoa(inp), SHORTLEN);
        fprintf(stderr, "Added criterion: %s; Value: %lu; Max value %lu; Svalue: %s; Family: %i; Addr: %s; Netmask: %s;"
#if HAVE_INET_NTOP
          " In6_addr: %s; In6_bits: %u"
#endif
          "\n",
            key, addl->value, addl->max_value, addl->svalue, addl->family, addr, mask
#if HAVE_INET_NTOP
            ,inet_ntop(AF_INET6, addl->in6_addr.s6_addr, buf, INET6_ADDRSTRLEN), addl->in6_bits
#endif
        );
      }

      addl->next = selection;
      selection = addl;

      selection_count++;

    }

  }

  free(parms);

  if (seen) {
    free(sel);
  }

  return;
}

/* Adds an item to the sort list.                         *
 * If sort_order->position = -35768, then only order flag *
 *   was present and entry should not be used for sorting */
struct sort_order * add_sort(struct sort_order *new, char *key, char *parm, int linenum, int mode)
{
  struct sort_order *srt = NULL, *have = NULL, *last = NULL, *this = NULL, *prev = NULL, *next = NULL, *list = NULL, *old = NULL;
  int seen, j;

  for (j=0; keys[j].key != NULL; j++) {
    if(strcmp(keys[j].key, key) == 0)
      break;
  }

  list = new;

  if (keys[j].key == NULL) {
    if (opt.verbose >= VERBOSE_ALERT)
      fprintf(stderr, "Skipping unrecognized sort field: %s\n", strip_nl(key));

    return list;
  }

  if (!keys[j].sort_key) {
    if (opt.verbose >= VERBOSE_ERROR)
      fprintf(stderr, "ignoring key = %s\n", key);

    return list;
  }

  srt = xmalloc(sizeof(struct sort_order));
  srt->next = NULL;
  xstrncpy(srt->keyname, key, SHORTLEN);
  srt->order = 0;
  srt->position = 0;
  srt->field = keys[j].sort_key;

  switch (mode) {
    case DEF_SORT:
      srt->position = num_parameter(parm, linenum);
      opt.use_geoip |= keys[j].use_geoip;
      break;
    case DEF_ORDER:
      if (yes_or_no(parm, linenum)){
        srt->order = ORDER_DESCENDING;
      }
      break;
    default:
      /* error */
      break;
  }

  have = list;
  seen = 0;
  while (have != NULL) {
    if (have->field == srt->field) {
      seen = 1;
      break;
    } else {
      prev = have;
      have = have->next;
    }
  }

  if (seen) {
    if (mode == DEF_SORT) {
      /* If we have seen key before and are doing sort position now,
       * break it out of list and treat it as new item to be inserted */
      have->position = srt->position;
        /* if item is currently first in list, make second item first */
        if (list == have) list = list->next;
        old = srt;
        srt = have;
        free(old);
        if (prev != NULL) {
          prev->next = have->next;
        }
        srt->next = NULL;
    } else {
      if (mode == DEF_ORDER) {
        have->order = srt->order;
      } else {;} /* error */ 
    }
  } else if (mode == DEF_ORDER) {
    srt->position = SORT_INV_NO_SORT;
    srt->next = list;
    list = srt; 
    free(have);
  }

  if (mode == DEF_SORT) {
    sort_count++;
    last = list;
    this = list;
    next = NULL;
    if (this != NULL) next = this->next;

    while (this != NULL) {
      if (opt.verbose >= VERBOSE_ERROR) {
        fprintf(stderr, "Inserting sort field %s into list: this->position = %i\n", key, this->position);
      }
      if (srt->position < this->position) {
        last = this;
        this = last->next;
        if (this != NULL) next = this->next;
        continue;
      } else {
        break;
      }
    }

    if (opt.verbose >= VERBOSE_DEBUG) {
      if (last != NULL) fprintf(stderr, "Last->position: %i\n", last->position);
      if (this != NULL) fprintf(stderr, "This->position: %i\n", this->position);
      if (next != NULL) fprintf(stderr, "Next->position: %i\n", next->position);
    }

    if (list == NULL || (list->position == SORT_INV_NO_SORT && last->position == SORT_INV_NO_SORT)) {
      if (opt.verbose >= VERBOSE_DEBUG) {
        fprintf(stderr, "First item seen; position: %i\n", srt->position);
      }
      if (list != NULL)
        srt->next = list; 
      list = srt;
    } else { /* not first item seen */
      if (list->next == NULL || (this == NULL && list->next->position == SORT_INV_NO_SORT \
           && last != NULL && srt->position > last->position)) {
        if (opt.verbose >= VERBOSE_DEBUG) {
          fprintf(stderr, "Second item seen; position: %i\n", srt->position);
        }
        if (srt->position < list->position) {
          if (list->next != NULL)
            srt->next = list->next; 
          list->next = srt;
        } else {
          this = list;
          list = srt;
          list->next = this;
        }
      } else { /* not second item seen */
        if (opt.verbose >= VERBOSE_DEBUG) {
          fprintf(stderr, "Not first or second item; position: %i\n", srt->position);
        }
        if (this == NULL) {
          if (opt.verbose >= VERBOSE_DEBUG) {
            fprintf(stderr, "this->position = NULL = place item last\n");
          }
          last->next = srt;
          srt->next = NULL;
        } else { /* not last */
          if (this->position == last->position) {
            if (opt.verbose >= VERBOSE_DEBUG) {
              fprintf(stderr, "this->position = last->position = place item first\n");
            }
            srt->next = list;
            list = srt;
          } else { /* not first */
            if (opt.verbose >= VERBOSE_DEBUG) {
              fprintf(stderr, "this->position < last->position = insert item\n");
            }
            last->next = srt;
            srt->next = this;
          } /* end: else clause of test to place item first */
        } /* end: else clause of test to place item last */
      } /* end: else clause of test for second item seen */
    } /* end: else clause of test for first item seen */
  } /* end: if(mode == DEF_SORT} */

  if (opt.verbose >= VERBOSE_WARNING) {
    switch (mode) {
      case DEF_ORDER:
        fprintf(stderr, "Added descending flag for sort field %s\n", key);
        break;
      case DEF_SORT:
          fprintf(stderr, "Added sort field %s at position %i\n", key, srt->position);
        break;
      default:
        break;
    }
  }

  return list;
}

/* parses single line from reportdef */
/* called with (linebuf, linenum) */
void parse_reportdef(char *input, int linenum)
{
  char *command, *key, *parm, *ptr;
  unsigned int type, len = 0;

  while (*input == ' ' || *input == '\t')
    ++input;

  if (*input == '#' || *input == '\n')
    return;

  command = strdup(input);
  key = strdup(input);
  ptr = key;

  len = strcspn(command, " =");
  xstrncpy(key, command, len + 1);

  parm = strpbrk(command, "=");
  if (!(parm == NULL))
    parm++;

  type = DEF_NONE;

  if (strncmp(key, "w_", 2) == 0)
    type = DEF_WHERE;

  if (strncmp(key, "i_", 2) == 0)
    type = DEF_INVERT;

  if (strncmp(key, "s_", 2) == 0)
    type = DEF_SORT;

  if (strncmp(key, "o_", 2) == 0)
    type = DEF_ORDER;

  if (strncmp(key, "c_", 2) == 0)
    type = DEF_COLUMN;

/* Ignore these */
  if (strncmp(key, "h_", 2) == 0)
    type = DEF_NONE;

  if (strncmp(key, "l_", 2) == 0)
    type = DEF_NONE;

  if (type != DEF_NONE) {
     key++; key++;
  }

/* Other options */
  if (strncmp(key, "summarize", 9) == 0)
    type = DEF_SUM;

  if (strncmp(key, "populate_cache", 14) == 0) {
    opt.resolve_hosts |= CACHE_POPULATE;
    type = DEF_OPTION;
  }

  if (strncmp(key, "upd_hosts", 9) == 0) {
    opt.resolve_hosts |= CACHE_UPDATE;
    type = DEF_OPTION;
  }

  if (strncmp(key, "populate_cache_full", 19) == 0) {
    opt.resolve_hosts |= CACHE_POPULATE_FULL;
    type = DEF_OPTION;
  }

  if (strncmp(key, "syslog_dir", 10) == 0) {
    xstrncpy(opt.pathname, str_parameter(parm), SHORTLEN);
    type = DEF_OPTION;
  }

  if (strncmp(key, "db", 2) == 0) {
    xstrncpy(opt.db, str_parameter(parm), SHORTLEN);
    type = DEF_OPTION;
  }

  if (strncmp(key, "mysql_server", 12) == 0) {
    xstrncpy(opt.mysql_server, str_parameter(parm), SHORTLEN);
    type = DEF_OPTION;
  }

  if (strncmp(key, "mysql_wfwl_db", 13) == 0) {
    xstrncpy(opt.mysql_wfwl_db, str_parameter(parm), SHORTLEN);
    type = DEF_OPTION;
  }

  if (strncmp(key, "mysql_user", 10) == 0) {
    xstrncpy(opt.mysql_user, str_parameter(parm), SHORTLEN);
    type = DEF_OPTION;
  }

  if (strncmp(key, "mysql_pass", 10) == 0) {
    xstrncpy(opt.mysql_pass, str_parameter(parm), SHORTLEN);
    type = DEF_OPTION;
  }

  if (strncmp(key, "pgsql_server", 12) == 0) {
    xstrncpy(opt.pgsql_server, str_parameter(parm), SHORTLEN);
    type = DEF_OPTION;
  }

  if (strncmp(key, "pgsql_db", 8) == 0) {
    xstrncpy(opt.pgsql_db, str_parameter(parm), SHORTLEN);
    type = DEF_OPTION;
  }

  if (strncmp(key, "pgsql_user", 10) == 0) {
    xstrncpy(opt.pgsql_user, str_parameter(parm), SHORTLEN);
    type = DEF_OPTION;
  }

  if (strncmp(key, "pgsql_pass", 10) == 0) {
    xstrncpy(opt.pgsql_pass, str_parameter(parm), SHORTLEN);
    type = DEF_OPTION;
  }

  if (strncmp(key, "pgsql_wfwl_schema", 17) == 0) {
    xstrncpy(opt.pgsql_wfwl_schema, str_parameter(parm), SHORTLEN - 1);
    type = DEF_OPTION;
  }

  if (strncmp(key, "pgsql_have_namespace", 20) == 0) {
    opt.pgsql_have_namespace = 1;
    type = DEF_OPTION;
  }

  if (strncmp(key, "geoip_version", 13) == 0) {
    opt.geoip_version = num_parameter(parm, linenum);
    type = DEF_OPTION;
  }

  if (strncmp(key, "geoip2_locale", 11) == 0) {
    xstrncpy(opt.geoip2_locale, str_parameter(parm), BUFSIZE);
    type = DEF_OPTION;
  }

  if (strncmp(key, "geoip2_file", 11) == 0) {
    xstrncpy(opt.geoip2_file, str_parameter(parm), BUFSIZE);
    type = DEF_OPTION;
  }

  if (strncmp(key, "geoip_php_lookup", 16) == 0) {
    if(yes_or_no(parm, linenum))
	opt.use_geoip |= GEOIP_OFF;
    type = DEF_OPTION;
  }

  if (type == DEF_OPTION) {
    if (opt.verbose >= VERBOSE_ALERT) {
      if (strncmp(key,"mysql_pass", 10) == 0||strncmp(key,"pgsql_pass", 10) == 0) {
        fprintf(stderr, "Added option: %s; Value: %s\n", key, "************");
      } else {
        fprintf(stderr, "Added option: %s; Value: %s\n", key, strip_nl(str_parameter(parm)));
      }
    }
  } else if (type != DEF_NONE) {
    if (opt.verbose >= VERBOSE_ERROR) {
      fprintf(stderr, "Processing report definition line Type: %u; Key: %s; Parameter: %s", type, key, parm);
    }
  } else { /* type == DEF_NONE */ 
    if (opt.verbose >= VERBOSE_ERROR) {
      fprintf(stderr, "Ignoring report definition line: %s", command);
    }
  }

  if (!(parm == NULL))
  switch (type) {
    case DEF_SUM:
      if (yes_or_no(parm, linenum)) {
        opt.mode = LOG_SUMMARY;
      } else {
        opt.mode = LOG_DETAIL;
      }
      break;
    case DEF_WHERE:
      add_selection(key, parm, linenum, DEF_WHERE);
      break;
    case DEF_INVERT:
      add_selection(key, parm, linenum, DEF_INVERT);
      break;
    case DEF_SORT:
      sort = add_sort(sort, key, parm, linenum, DEF_SORT);
      break;
    case DEF_ORDER:
      sort = add_sort(sort, key, parm, linenum, DEF_ORDER);
      break;
    case DEF_COLUMN:
      fields = add_column(fields, key, parm, linenum);
      break;
    case DEF_OPTION:
    case DEF_NONE:
      break;
    default:
      /* error */
      break;
  }

  free(ptr);
  free(command);

  return;
}

/* Entry point; called from main.c */
unsigned char read_reportdef(char *reportdef)
{
  char buf[BUFSIZE], *name;
  FILE *fd;
  int linenum = 1, retval;
  char keyname[SHORTLEN];
  char position[SHORTLEN];
  char addr[16], mask[16];
  struct stat info;
  struct sort_order *srt;
  struct field_order *col;
  struct selection *sel;
  struct in_addr in;

  name = strdup(reportdef);

  /* validate reportdef file */
  if (strcmp(reportdef, "-")) {
    retval = stat(reportdef, &info);
    if (retval == -1) {
      fprintf(stderr, "Cannot open report definition file %s, exiting.\n", reportdef);
      return EXIT_FAILURE;
    }
    if (!S_ISREG(info.st_mode)) {
      fprintf(stderr, "%s is not a regular file, exiting.\n", reportdef);
      return EXIT_FAILURE;
    }

    if(opt.verbose >= VERBOSE_INFO)
      fprintf(stderr, "Processing report definition file '%s'\n", name);

    fd = fopen(name, "r");
    if (fd == NULL) {
      fprintf(stderr, "fopen %s: %s\n", name, strerror(errno));
      exit(EXIT_FAILURE);
    }
  } else {

    if(opt.verbose >= VERBOSE_INFO)
      fprintf(stderr, "Processing report definition from stdin\n");

    fd = stdin;
  }

  /* initialize fields_used */
  /* this is used for summarization testing.                                *
   * tests done against this static data are faster by a factor of at       *
   * least two over iterating though field_order linked list for each line. */
  fields_used.count = 0;
  fields_used.sname = 0;
  fields_used.local_time = 0;
  fields_used.earliest = 0;
  fields_used.latest = 0;
  fields_used.hostname = 0;
  fields_used.log_label = 0;
  fields_used.protocol = 0;
  fields_used.family = 0;
  fields_used.totlen = 0;
  fields_used.shost = 0;
  fields_used.sport = 0;
  fields_used.dhost = 0;
  fields_used.dport = 0;
  fields_used.shost_name = 0;
  fields_used.dhost_name = 0;
  fields_used.src_service = 0;
  fields_used.dst_service = 0;
  fields_used.flags = 0;
  fields_used.oob_protocol = 0;
  fields_used.mac_saddr_str = 0;
  fields_used.mac_daddr_str = 0;
  fields_used.raw_mac = 0;
  fields_used.fwmark = 0;
  fields_used.inif = 0;
  fields_used.outif = 0;
  fields_used.tos = 0;
  fields_used.ttl = 0;
  fields_used.ihl = 0;
  fields_used.csum = 0;
  fields_used.ipid = 0;
  fields_used.fragoff = 0;
  fields_used.pay_len = 0;
  fields_used.flowlabel = 0;
  fields_used.tcp_seq = 0;
  fields_used.tcp_ack_seq = 0;
  fields_used.tcp_window = 0;
  fields_used.tcp_urgp = 0;
  fields_used.udp_len = 0;
  fields_used.icmp_type = 0;
  fields_used.icmp_code = 0;
  fields_used.icmp_echoid = 0;
  fields_used.icmp_echoseq = 0;
  fields_used.icmp_gw = 0;
  fields_used.icmp_mtu = 0;
  fields_used.ahesp_spi = 0;
  fields_used.continent_code_s = 0;
  fields_used.continent_name_s = 0;
  fields_used.country_iso_code_s = 0;
  fields_used.country_name_s = 0;
  fields_used.subdivision_1_iso_code_s = 0;
  fields_used.subdivision_1_name_s = 0;
  fields_used.subdivision_2_iso_code_s = 0;
  fields_used.subdivision_2_name_s = 0;
  fields_used.city_name_s = 0;
  fields_used.metro_code_s = 0;
  fields_used.accuracy_radius_s = 0;
  fields_used.time_zone_s = 0;
  fields_used.registered_country_geoname_id_s = 0;
  fields_used.represented_country_geoname_id_s = 0;
  fields_used.is_anonymous_proxy_s = 0;
  fields_used.is_satellite_provider_s = 0;
  fields_used.postal_code_s = 0;
  fields_used.latitude_s = 0;
  fields_used.longitude_s = 0;
  fields_used.continent_code_d = 0;
  fields_used.continent_name_d = 0;
  fields_used.country_iso_code_d = 0;
  fields_used.country_name_d = 0;
  fields_used.subdivision_1_iso_code_d = 0;
  fields_used.subdivision_1_name_d = 0;
  fields_used.subdivision_2_iso_code_d = 0;
  fields_used.subdivision_2_name_d = 0;
  fields_used.city_name_d = 0;
  fields_used.metro_code_d = 0;
  fields_used.accuracy_radius_d = 0;
  fields_used.time_zone_d = 0;
  fields_used.registered_country_geoname_id_d = 0;
  fields_used.represented_country_geoname_id_d = 0;
  fields_used.is_anonymous_proxy_d = 0;
  fields_used.is_satellite_provider_d = 0;
  fields_used.postal_code_d = 0;
  fields_used.latitude_d = 0;
  fields_used.longitude_d = 0;

  /* initialize select_sum */
  select_sum.min_count = 0;
  select_sum.max_count = 0;
  select_sum.max_earliest = 0;
  select_sum.min_latest = 0;

  while (fgets(buf, BUFSIZE, fd)) {
    if (opt.verbose >= VERBOSE_DEBUG)
      fprintf(stderr, "Input line is: %s", buf);

    parse_reportdef(buf, linenum);
    linenum++;
  }

  if (opt.mode == LOG_DETAIL && fields_used.sname == 0) {
    strcpy(keyname, "sname");
    strcpy(position, "99");
    fields = add_column(fields, keyname, position, 999);
  }

  if (opt.verbose >= VERBOSE_INFO) {
    fprintf(stderr, "Number of columns: %u\n", column_count);
  }

  if (opt.verbose >= VERBOSE_NOTICE) {
    col = fields;
    while (col != NULL) {
      fprintf(stderr, "COLUMNS: Key: %s; Field %i; Position: %i\n", col->keyname, col->field, col->position);
      col = col->next;
    }
  }

  if (opt.verbose >= VERBOSE_INFO) {
    fprintf(stderr, "Number of criteria: %u\n", selection_count);
  }

  if (opt.verbose >= VERBOSE_NOTICE) {
    sel = selection;
    while (sel != NULL) {
      in.s_addr = htonl(sel->in_addr);
      xstrncpy(addr, inet_ntoa(in), SHORTLEN);
      in.s_addr = htonl(sel->netmask);
      xstrncpy(mask, inet_ntoa(in), SHORTLEN);
      fprintf(stderr, "SELECTION: Key: %s; Field %i; Value: %lu Max value: %lu; Double: %.7g; Max double: %.7g; Svalue: %s; Family: %i; Addr: %s; Netmask: %s;"
#if HAVE_INET_NTOP
        " In6_addr: %s; In6_bits: %u;"
#endif
        " Invert: %i; Have Value: %i\n",
        sel->keyname, sel->field, sel->value, sel->max_value, sel->double_value, sel->max_double_value, sel->svalue, sel->family, addr, mask,
#if HAVE_INET_NTOP
        inet_ntop(AF_INET6, sel->in6_addr.s6_addr, buf, INET6_ADDRSTRLEN), sel->in6_bits,
#endif
        sel->invert, sel->have_value);

      sel = sel->next;
    }
    if (select_sum.min_count !=0)
      fprintf(stderr, "SELECTION: Key: min_count; Value: %lu\n", select_sum.min_count);
    if (select_sum.max_count !=0)
      fprintf(stderr, "SELECTION: Key: max_count; Value: %lu\n", select_sum.max_count);
    if (select_sum.max_earliest !=0)
      fprintf(stderr, "SELECTION: Key: max_earliest; Value: %lu\n", (unsigned long int)select_sum.max_earliest);
    if (select_sum.min_latest !=0)
      fprintf(stderr, "SELECTION: Key: min_latest; Value: %lu\n", (unsigned long int)select_sum.min_latest);
  }

  if (opt.verbose >= VERBOSE_INFO) {
    fprintf(stderr, "Number of sorts: %u\n", sort_count);
  }

  if (opt.verbose >= VERBOSE_NOTICE) {
    srt = sort;
    while (srt != NULL) {
      fprintf(stderr, "SORT: Key: %s; Field %i; Position: %i Order: %i\n",
        srt->keyname, srt->field, srt->position, srt->order);
      srt = srt->next;
    }
  }

  if ((strcmp(reportdef, "-"))) { 
    if(opt.verbose >= VERBOSE_WARNING)
      fprintf(stderr, "Done processing report definition file '%s'\n", name);
  } else {
    if(opt.verbose >= VERBOSE_WARNING)
      fprintf(stderr, "Done processing report definition from stdin\n");
  }

  free(name);

  retval = fclose(fd);
  if (retval == EOF) {
    perror("fclose");
    exit(EXIT_FAILURE);
  }

  return EXIT_SUCCESS;
}
